from contextlib import AsyncExitStack, asynccontextmanager
from typing import AsyncGenerator, Dict, List, TypeVar

from ._mocked import Mocked

__all__ = ("stacked",)

MockedType = TypeVar("MockedType", List[Mocked], Dict[str, Mocked])


@asynccontextmanager
async def stacked(mocks: MockedType) -> AsyncGenerator[MockedType, None]:
    """
    Stack multiple mocks into a single context manager for easier management.

    This context manager allows you to stack multiple mocks (either in a list or a dictionary)
    into a single context, ensuring they are all activated upon entering and deactivated
    upon exiting. It simplifies the process of managing multiple mocks simultaneously.

    :param mocks: A list or dictionary of `Mocked` objects to be activated in the context.
                  If a list is provided, the mocks are returned as a list; if a dictionary is
                  provided, they are returned as a dictionary with the same keys.
    :return: Yields the activated mocks. If the input was a list, it yields a list of active mocks;
             if the input was a dictionary, it yields a dictionary of active mocks
             keyed by the same names.
    :raises TypeError: If the `mocks` parameter is not a list or a dictionary.
    """
    async with AsyncExitStack() as stack:
        if isinstance(mocks, list):
            yield [await stack.enter_async_context(mock) for mock in mocks]
        elif isinstance(mocks, dict):
            yield {name: await stack.enter_async_context(mock) for name, mock in mocks.items()}
        else:
            raise TypeError(f"Unsupported type: {type(mocks)}")



import logging
import os

from ._filter import Filter
from ._logger import Logger
from ._request_filter import RequestFilter
from ._system_log_filter import SystemLogFilter
from .formatters import Formatter, SimpleFormatter, TemplateFormatter

__all__ = ("Logger", "Filter", "Formatter", "SimpleFormatter", "TemplateFormatter",
           "SystemLogFilter", "RequestFilter",)

# Set custom Logger class

_old_logger_class = logging.getLoggerClass()
logging.setLoggerClass(Logger)

# Register Loggers

logger: Logger = logging.getLogger("jj.access_logger")  # type: ignore
log_level = getattr(logging, os.getenv("JJ_LOG_LEVEL", "INFO"))
logger.setLevel(log_level)
logger.propagate = False

# Register Filter

filter_ = Filter()
request_filter_ = RequestFilter()
logger.addFilter(filter_)
logger.addFilter(request_filter_)
if log_level != logging.DEBUG:
    logger.addFilter(SystemLogFilter())

# Register Handler

handler = logging.StreamHandler()
logger.addHandler(handler)

# Register Formatter

formatter = TemplateFormatter()
handler.setFormatter(formatter)

# Restore default Logger class

logging.setLoggerClass(_old_logger_class)



from datetime import datetime
from typing import List, Optional

from ...requests import Request
from ...responses import StreamResponse
from ._history_item import HistoryItem
from ._history_request import HistoryRequest
from ._history_response import HistoryResponse

__all__ = ("HistoryRepository",)


class HistoryRepository:
    def __init__(self) -> None:
        self._storage: List[HistoryItem] = []

    async def add(self,
                  request: Request,
                  response: StreamResponse,
                  tags: Optional[List[str]] = None,
                  created_at: Optional[datetime] = None) -> None:
        req = await HistoryRequest.from_request(request)
        res = await HistoryResponse.from_response(response)
        self._storage.insert(0, {
            "request": req,
            "response": res,
            "tags": tags or [],
            "created_at": created_at or datetime.utcnow(),
        })

    async def delete_by_tag(self, tag: str) -> None:
        self._storage = [x for x in self._storage if tag not in x["tags"]]

    async def get_by_tag(self, tag: str) -> List[HistoryItem]:
        return [x for x in self._storage if tag in x["tags"]]

    async def clear(self) -> None:
        self._storage.clear()



from typing import Any, Callable, Optional, TypeVar, Union

from jj.expiration_policy import ExpirationPolicy
from jj.matchers import LogicalMatcher, RequestMatcher

from ._create_remote_handler import create_remote_handler
from ._history import (
    HistoryAdapterType,
    HistoryFormatter,
    PrettyHistoryFormatter,
    default_history_adapter,
)
from ._mocked import Mocked
from ._remote_response import RemoteResponseType
from ._settings import (
    get_remote_mock_disposable,
    get_remote_mock_pprint,
    get_remote_mock_pprint_length,
    get_remote_mock_pprint_width,
)

__all__ = ("MockedFactory",)

F = TypeVar("F", bound=Callable[..., Any])


class MockedFactory:
    """
    Factory for creating `Mocked` instances with remote handlers.

    This class provides multiple ways to create and use mocked HTTP requests:
    - Direct instantiation via `__call__`
    - As a decorator with mock injection via `with_mock`

    The factory handles configuration from environment variables and sets up
    the remote handler with the specified matchers and responses.
    """

    def __call__(
        self,
        matcher: Union[RequestMatcher, LogicalMatcher],
        response: RemoteResponseType,
        expiration_policy: Optional[ExpirationPolicy] = None,
        *,
        disposable: Optional[bool] = None,
        prefetch_history: bool = True,
        history_formatter: Optional[HistoryFormatter] = None,
        history_adapter: Optional[HistoryAdapterType] = default_history_adapter
    ) -> Mocked:
        """
        Create a `Mocked` instance for intercepting and mocking HTTP requests.

        This method can be used as a context manager or decorator to mock HTTP requests
        that match the specified criteria.

        :param matcher: A matcher that determines which requests to intercept.
        :param response: The response to return for matched requests.
        :param expiration_policy: Optional policy that determines when the mock expires.
        :param disposable: Whether the mock should be automatically deregistered after use.
                          If `None`, uses the value from `JJ_REMOTE_MOCK_DISPOSABLE` env var.
        :param prefetch_history: Whether to automatically fetch request history when exiting
                                the context. Defaults to `True`.
        :param history_formatter: Optional formatter for displaying request/response history.
                                 If `None` and `JJ_REMOTE_MOCK_PPRINT` is "true", uses
                                 `PrettyHistoryFormatter`.
        :param history_adapter: Optional adapter for handling request/response history.
                               Defaults to `default_history_adapter`.
        :return: A configured `Mocked` instance.
        """
        return self._mocked(matcher, response, expiration_policy,
                            disposable=disposable,
                            prefetch_history=prefetch_history,
                            history_formatter=history_formatter,
                            history_adapter=history_adapter)

    def with_mock(
        self,
        matcher: Union[RequestMatcher, LogicalMatcher],
        response: RemoteResponseType,
        expiration_policy: Optional[ExpirationPolicy] = None,
        *,
        disposable: Optional[bool] = None,
        prefetch_history: bool = True,
        history_formatter: Optional[HistoryFormatter] = None,
        history_adapter: Optional[HistoryAdapterType] = default_history_adapter
    ) -> Callable[[F], F]:
        """
        Create a decorator that injects a `Mocked` instance into the decorated function.

        This method returns a decorator that passes the mock as the first argument to
        the decorated function, allowing direct access to the mock instance within tests.

        :param matcher: A matcher that determines which requests to intercept.
        :param response: The response to return for matched requests.
        :param expiration_policy: Optional policy that determines when the mock expires.
        :param disposable: Whether the mock should be automatically deregistered after use.
                          If `None`, uses the value from `JJ_REMOTE_MOCK_DISPOSABLE` env var.
        :param prefetch_history: Whether to automatically fetch request history when exiting
                                the context. Defaults to `True`.
        :param history_formatter: Optional formatter for displaying request/response history.
                                 If `None` and `JJ_REMOTE_MOCK_PPRINT` is "true", uses
                                 `PrettyHistoryFormatter`.
        :param history_adapter: Optional adapter for handling request/response history.
                               Defaults to `default_history_adapter`.
        :return: A decorator that injects the mock into the decorated function.
        """
        return self._mocked(matcher, response, expiration_policy,
                            disposable=disposable,
                            prefetch_history=prefetch_history,
                            history_formatter=history_formatter,
                            history_adapter=history_adapter).with_mock

    def _mocked(
        self,
        matcher: Union[RequestMatcher, LogicalMatcher],
        response: RemoteResponseType,
        expiration_policy: Optional[ExpirationPolicy] = None,
        *,
        disposable: Optional[bool] = None,
        prefetch_history: bool = True,
        history_formatter: Optional[HistoryFormatter] = None,
        history_adapter: Optional[HistoryAdapterType] = default_history_adapter
    ) -> Mocked:
        if disposable is None:
            disposable = get_remote_mock_disposable()

        if (history_formatter is None) and get_remote_mock_pprint():
            history_formatter = PrettyHistoryFormatter(
                get_remote_mock_pprint_width(),
                get_remote_mock_pprint_length()
            )

        handler = create_remote_handler(matcher, response, expiration_policy,
                                        history_adapter=history_adapter)

        return Mocked(handler, disposable=disposable, prefetch_history=prefetch_history,
                      history_formatter=history_formatter)



import json
from typing import Any, Dict, Optional, Union

from aiohttp import web
from aiohttp.web_request import FileField
from multidict import MultiDictProxy, MultiMapping

from ..responses import StreamResponse

__all__ = ("Request",)


class Request(web.Request):
    """
    Extends aiohttp's `web.Request` to add support for URL path segments and custom parameters.

    This class enhances the base `web.Request` by adding a `segments` attribute for handling
    dynamic URL path segments and provides a shorthand property for accessing query parameters.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the Request object.

        :param args: Positional arguments to pass to the base `web.Request`.
        :param kwargs: Keyword arguments to pass to the base `web.Request`.
        """
        super().__init__(*args, **kwargs)
        self._segments: Optional[Dict[str, str]] = None

    @property
    def params(self) -> "MultiMapping[str]":
        """
        Get the query parameters of the request.

        This is a shorthand property that returns the same result as `self.query`.

        :return: A `MultiMapping` object containing the query parameters.
        """
        return self.query

    @property
    def segments(self) -> Dict[str, str]:
        """
        Get the URL path segments.

        This property holds dynamic URL segments that can be set and retrieved for the request.
        If no segments are set, an empty dictionary is returned.

        :return: A dictionary containing the path segments of the request,
                 or an empty dictionary if not set.
        """
        if self._segments is None:
            return {}
        return self._segments

    @segments.setter
    def segments(self, segments: Optional[Dict[str, str]]) -> None:
        """
        Set the URL path segments.

        This allows setting a dictionary of path segments for the request. Path segments
        are often used when dealing with dynamic URL paths that contain variable placeholders.

        :param segments: A dictionary of path segments, or `None` to reset the segments.
        """
        self._segments = segments

    @property
    def raw_data(self) -> Union[bytes, None]:
        """
        Get the raw request body as bytes.

        :return: The raw request body as bytes, or `None` if not available.
        """
        return self._read_bytes

    @property
    def post_data(self) -> Union[MultiDictProxy[Union[str, bytes, FileField]], None]:
        """
        Get the parsed form data from the request.

        This property provides access to the form-encoded body of the request if available.

        :return: A `MultiDictProxy` containing the form data, or `None` if not available.
        """
        return self._post

    @property
    def json_data(self) -> Union[Any, None]:
        """
        Get the request body as a parsed JSON object.

        This property decodes the request body using the specified charset (defaulting to UTF-8)
        and parses it as JSON.

        :return: The parsed JSON data, or `None` if the request body is empty.
        :raises json.JSONDecodeError: If the request body is not valid JSON.
        """
        if self._read_bytes is None:
            return None
        encoding = self.charset or "utf-8"
        return json.loads(self._read_bytes.decode(encoding))

    async def _prepare_hook(self, response: StreamResponse) -> None:  # type: ignore
        """
        Prepare the request before it is processed by the server.

        This hook method can be used to perform any custom preparation or pre-processing
        on the request object before it is sent for handling.

        :param response: The response object associated with the request.
        """
        pass



from logging import LogRecord

from jj.logs import Filter
from jj.requests import Request
from jj.responses import Response

__all__ = ("SystemLogFilter",)


class SystemLogFilter(Filter):
    def filter_request(self, request: Request, record: LogRecord) -> bool:
        return "x-jj-remote-mock" not in request.headers

    def filter_response(self, response: Response, request: Request, record: LogRecord) -> bool:
        return self.filter_request(request, record)



import mimetypes

from ._delayed_response import DelayedResponse
from ._relay_response import RelayResponse
from ._response import Response
from ._static_response import StaticResponse
from ._stream_response import StreamResponse
from ._template_response import TemplateResponse

__all__ = ("Response", "StaticResponse", "StreamResponse", "RelayResponse",
           "DelayedResponse", "TemplateResponse",)


mimetypes.add_type("application/json", ".json")



from typing import Any, Dict, Tuple
from urllib.parse import urljoin

from aiohttp import ClientSession
from aiohttp.web_request import BaseRequest
from multidict import CIMultiDict, CIMultiDictProxy, MultiMapping
from packed import packable

from ._response import Response

__all__ = ("RelayResponse",)


# http://tools.ietf.org/html/rfc2616#section-13.5.1
_HOP_BY_HOP_HEADERS = (
    "connection",
    "keep-alive",
    "proxy-authenticate",
    "proxy-authorization",
    "te",
    "trailers",
    "transfer-encoding",
    "upgrade")

_FILTERED_HEADERS = _HOP_BY_HOP_HEADERS + ("host", "content-length")

_TargetResponseType = Tuple[int, Any, CIMultiDictProxy[str], bytes]


@packable("jj.responses.RelayResponse")
class RelayResponse(Response):
    """
    Relays the request to a target server and returns the response.

    The `RelayResponse` class is a special kind of response that forwards incoming
    requests to a target server, retrieves the response from the target, and relays
    it back to the client.
    """

    def __init__(self, *, target: str) -> None:
        """
        Initialize the RelayResponse with a target URL.

        :param target: The base URL of the target server to which the request will be relayed.
        """
        super().__init__()
        self._target = target
        self._prepare_hook_called = False

    @property
    def target(self) -> str:
        """
        Get the target URL for the relayed request.

        :return: The base URL of the target server.
        """
        return self._target

    def _filter_headers(self, headers: MultiMapping[str]) -> CIMultiDict[str]:
        """
        Filter headers to exclude hop-by-hop and other headers not forwarded.

        This method removes headers that should not be forwarded to the target server,
        such as hop-by-hop headers (per RFC 2616, Section 13.5.1) and certain others
        like `host` and `content-length`.

        :param headers: The original headers of the incoming request.
        :return: A new `CIMultiDict` containing only the headers that should be forwarded.
        """
        filtered: CIMultiDict[str] = CIMultiDict()
        for key, value in headers.items():
            if key.lower() not in _FILTERED_HEADERS:
                filtered[key] = value
        return filtered

    async def _do_target_request(self, request: BaseRequest) -> _TargetResponseType:
        """
        Perform the request to the target server and retrieve the response.

        This method constructs a new request based on the incoming request, forwards it to
        the target server, and retrieves the response. It also handles reading the response
        body and returning necessary response details.

        :param request: The incoming client request to be relayed.
        :return: A tuple containing the status code, reason, filtered headers,
                 and the body of the target response.
        """
        url = urljoin(self._target, request.path)
        headers = self._filter_headers(request.headers)

        data = await request.read()

        async with ClientSession(auto_decompress=False) as session:
            async with session.request(request.method, url, params=request.query, headers=headers,
                                       data=data, allow_redirects=True) as response:
                body = await response.read()
        return response.status, response.reason, response.headers, body

    async def _prepare_hook(self, request: BaseRequest) -> "RelayResponse":
        """
        Prepare the relay response by performing the request to the target server.

        This method is called during the response preparation phase and ensures that
        the request is sent to the target server. It also processes the response from
        the target, including setting the appropriate status, headers, and body.

        :param request: The incoming request to be relayed.
        :return: The `RelayResponse` object after the target request is processed.
        """
        if self._prepare_hook_called:
            return self
        status, reason, headers, body = await self._do_target_request(request)
        self.set_status(status, reason)
        self._headers = CIMultiDict(self._filter_headers(headers))
        self.body = body
        self._prepare_hook_called = True
        return self

    def copy(self) -> "RelayResponse":
        """
        Create a copy of the RelayResponse object.

        This method clones the current `RelayResponse` object, ensuring that the target URL
        is preserved.

        :return: A new instance of `RelayResponse` with the same target URL.
        :raises AssertionError: If the response is already prepared.
        """
        assert not self.prepared
        return self.__class__(target=self._target)

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the RelayResponse object for serialization.

        This method packs the target URL into a dictionary, which can then be used for
        serialization or transmission.

        :return: A dictionary containing the target URL.
        """
        return {"target": self._target}

    @classmethod
    def __unpacked__(cls, *, target: str, **kwargs: Any) -> "RelayResponse":  # type: ignore
        """
        Reconstruct a RelayResponse instance from unpacked parameters.

        :param target: The target URL to relay requests to.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new `RelayResponse` instance with the specified target.
        """
        return cls(target=target)



from logging import LogRecord
from os import linesep

from ...requests import Request
from ...responses import Response
from ._formatter import Formatter

__all__ = ("SimpleFormatter",)


class SimpleFormatter(Formatter):
    def format_request(self, request: Request, record: LogRecord) -> str:
        return "-> {} {}".format(request.method, request.url.path)

    def format_response(self, response: Response, request: Request, record: LogRecord) -> str:
        res = linesep.join([
            "-> {} {}".format(request.method, request.url.path),
            "<- {} {}".format(response.status, response.reason),
            "",
        ])
        return res



import os

from jj.logs import SystemLogFilter  # backward compatibility

from ._create_remote_handler import create_remote_handler
from ._history import (
    HistoryAdapterType,
    HistoryFormatter,
    HistoryItem,
    HistoryRepository,
    HistoryRequest,
    HistoryResponse,
    PrettyHistoryFormatter,
    default_history_adapter,
)
from ._mock import Mock
from ._mocked import Mocked
from ._mocked_factory import MockedFactory
from ._remote_handler import RemoteHandler
from ._remote_mock import RemoteMock
from ._remote_response import RemoteResponseType
from ._settings import get_remote_mock_disposable, get_remote_mock_url
from ._stacked import stacked

__all__ = ("mocked", "stacked", "Mock", "Mocked", "create_remote_handler", "RemoteMock",
           "RemoteHandler", "HistoryRepository", "HistoryRequest", "HistoryResponse",
           "HistoryItem", "HistoryFormatter", "PrettyHistoryFormatter", "RemoteResponseType",
           "HistoryAdapterType", "SystemLogFilter", "default_history_adapter",
           "REMOTE_MOCK_URL", "REMOTE_MOCK_DISPOSABLE",
           "get_remote_mock_url", "get_remote_mock_disposable",)


# Deprecated: Use getter methods instead to avoid race conditions with runtime env loading
# like load_dotenv()
REMOTE_MOCK_URL = os.environ.get("JJ_REMOTE_MOCK_URL", "http://localhost:8080")
REMOTE_MOCK_DISPOSABLE = os.environ.get("JJ_REMOTE_MOCK_DISPOSABLE", "True")

REMOTE_MOCK_PPRINT = os.environ.get("JJ_REMOTE_MOCK_PPRINT", "False")
REMOTE_MOCK_PPRINT_LENGTH = os.environ.get("JJ_REMOTE_MOCK_PPRINT_LENGTH", "")
REMOTE_MOCK_PPRINT_WIDTH = os.environ.get("JJ_REMOTE_MOCK_PPRINT_WIDTH", "")

# backward compatibility
_REMOTE_MOCK_URL = REMOTE_MOCK_URL
_REMOTE_MOCK_DISPOSABLE = REMOTE_MOCK_DISPOSABLE


mocked = MockedFactory()



import argparse
from os import environ as env
from pathlib import Path

import jj
from jj.mock import Mock

from ._load_module import load_module


def run() -> None:
    """
    Run the remote HTTP mock server.

    This function sets up an argument parser to configure the server's host, port,
    and optionally load custom matcher modules. It retrieves values from the
    environment or provided arguments and starts the `jj` mock server.
    """
    parser = argparse.ArgumentParser("jj", description="Remote HTTP Mock Server")
    parser.add_argument(
        "-H", "--host", type=str, default=None,
        help="Host address to serve on (default: None)"
    )
    parser.add_argument(
        "-p", "--port", type=int, default=env.get("PORT", 8080),
        help="Port number to serve on (default: 8080)"
    )
    parser.add_argument(
        "--use-matchers", type=str, nargs="+",
        help="Path(s) to custom matcher module(s), separated by space."
    )
    args = parser.parse_args()

    if args.use_matchers:
        for module_path_str in args.use_matchers:
            module_path = Path(module_path_str)
            load_module(module_path)

    jj.serve(Mock(), host=args.host, port=args.port)



from asyncio import iscoroutinefunction
from functools import wraps
from types import TracebackType
from typing import Any, Callable, Generator, List, Optional, Type, TypeVar, Union, cast

from rtry import CancelledError, retry
from rtry.types import AttemptValue, DelayCallable, DelayValue, LoggerCallable, TimeoutValue

from ._history import HistoryFormatter, HistoryItem
from ._remote_handler import RemoteHandler
from ._utils import run_async

__all__ = ("Mocked",)

F = TypeVar("F", bound=Callable[..., Any])


class Mocked:
    """
    A context manager and decorator for mocking HTTP requests with remote handlers.

    This class provides multiple ways to mock HTTP requests:
    - As a context manager (sync or async) for scoped mocking
    - As a decorator to wrap entire functions with mocking
    - With the `with_mock` decorator to inject the mock instance into functions

    The class handles automatic registration/deregistration of mocks and can
    optionally fetch request history after use.
    """

    def __init__(self, handler: RemoteHandler, *,
                 disposable: bool = True,
                 prefetch_history: bool = True,
                 history_formatter: Optional[HistoryFormatter] = None
                 ) -> None:
        """
        Initialize a Mocked instance with a remote handler.

        :param handler: The remote handler that manages the mock registration and requests.
        :param disposable: Whether to automatically deregister the mock after use.
                          Defaults to `True`. Deprecated, will be removed in future versions.
        :param prefetch_history: Whether to automatically fetch request history when
                                exiting the context. Defaults to `True`.
        :param history_formatter: Optional formatter for displaying request/response history.
        """
        self._handler = handler
        self._disposable = disposable  # Deprecated, will be removed in future versions
        self._prefetch_history = prefetch_history  # Enable history prefetching in a `with` context
        self._history: Union[List[HistoryItem], None] = None
        self._history_formatter = history_formatter

    @property
    def handler(self) -> RemoteHandler:
        """
        Get the remote handler instance.

        :return: The remote handler managing this mock.
        """
        return self._handler

    @property
    def disposable(self) -> bool:
        """
        Check if the mock is disposable.

        :return: `True` if the mock will be automatically deregistered after use.
        """
        return self._disposable

    @property
    def prefetch_history(self) -> bool:
        """
        Check if history prefetching is enabled.

        :return: `True` if history will be automatically fetched when exiting context.
        """
        return self._prefetch_history

    @property
    def history(self) -> Union[List[HistoryItem], None]:
        """
        Get the fetched request/response history.

        :return: List of history items if fetched, otherwise `None`.
        """
        return self._history

    async def fetch_history(self) -> List[HistoryItem]:
        """
        Fetch the request/response history from the remote handler.

        This method retrieves all recorded interactions with the mock and stores
        them internally for later access via the `history` property.

        :return: List of history items containing request/response pairs.
        """
        self._history = await self._handler.fetch_history()
        return self._history

    async def wait_for_requests(self, count: int = 1, *,
                                timeout: TimeoutValue = 0,
                                attempts: Optional[AttemptValue] = None,
                                delay: Optional[Union[DelayValue, DelayCallable]] = None,
                                logger: Optional[LoggerCallable] = None) -> None:
        """
        Wait for a specified number of requests to be received by the mock.

        This method polls the history until the expected number of requests have been
        received or a timeout/retry limit is reached. Useful for testing async code
        that makes HTTP requests.

        :param count: The minimum number of requests to wait for. Defaults to 1.
        :param timeout: Maximum time to wait in seconds. 0 means no timeout.
        :param attempts: Maximum number of retry attempts. `None` means unlimited.
        :param delay: Delay between retry attempts. Can be a fixed value or callable.
        :param logger: Optional logger callable for retry attempts.
        """
        try:
            await retry(until=lambda x: len(x) < count,
                        attempts=attempts,
                        timeout=timeout,
                        delay=delay,
                        logger=logger)(self.fetch_history)()
        except CancelledError:
            pass

    def __call__(self, func: F) -> F:
        """
        Use the Mocked instance as a decorator.

        When used as a decorator, the mock will be active for the duration of
        the decorated function's execution.

        :param func: The function to wrap with mocking.
        :return: A wrapped function that executes with the mock active.
        :raises TypeError: If applied to a non-callable object.
        """
        if not callable(func):
            raise TypeError(
                f"mocked decorator can only be applied to functions or methods, "
                f"got {type(func).__name__}"
            )

        if iscoroutinefunction(func):
            @wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                async with self:
                    return await func(*args, **kwargs)

            return cast(F, async_wrapper)
        else:
            @wraps(func)
            def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
                with self:
                    return func(*args, **kwargs)

            return cast(F, sync_wrapper)

    def with_mock(self, func: F) -> F:
        """
        Decorator that injects the mock instance as the first argument to the function.

        This allows the decorated function to access the mock instance directly,
        useful for checking history or performing other mock operations.

        :param func: The function to wrap and inject the mock into.
        :return: A wrapped function that receives the mock as its first argument.
        :raises TypeError: If applied to a non-callable object.
        """
        if not callable(func):
            raise TypeError(
                f"mocked.with_mock decorator can only be applied to functions or methods, "
                f"got {type(func).__name__}"
            )

        if iscoroutinefunction(func):
            @wraps(func)
            async def async_wrapper() -> Any:
                async with self as mock:
                    return await func(mock)

            return cast(F, async_wrapper)
        else:
            @wraps(func)
            def sync_wrapper() -> Any:
                with self as mock:
                    return func(mock)

            return cast(F, sync_wrapper)

    def __await__(self) -> Generator[Any, None, "Mocked"]:
        """
        Support for awaiting the mock registration.

        :return: Generator that yields from handler registration.
        """
        yield from self._handler.register().__await__()
        return self

    async def __aenter__(self) -> "Mocked":
        """
        Async context manager entry.

        Registers the mock with the remote handler and clears any existing history.

        :return: The Mocked instance.
        """
        self._history = None
        await self._handler.register()
        return self

    async def __aexit__(self,
                        exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> None:
        """
        Async context manager exit.

        Optionally fetches history and deregisters the mock based on configuration.

        :param exc_type: Exception type if an exception was raised.
        :param exc_val: Exception value if an exception was raised.
        :param exc_tb: Exception traceback if an exception was raised.
        """
        if self._prefetch_history:
            await self.fetch_history()

        if self._disposable:
            await self._handler.deregister()

    def __enter__(self) -> "Mocked":
        """
        Sync context manager entry.

        Registers the mock with the remote handler using async-to-sync conversion.

        :return: The Mocked instance.
        """
        return run_async(self.__aenter__)

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_val: Optional[BaseException],
                 exc_tb: Optional[TracebackType]) -> None:
        """
        Sync context manager exit.

        Handles cleanup using async-to-sync conversion.

        :param exc_type: Exception type if an exception was raised.
        :param exc_val: Exception value if an exception was raised.
        :param exc_tb: Exception traceback if an exception was raised.
        """
        return run_async(self.__aexit__, exc_type, exc_val, exc_tb)

    def __repr__(self) -> str:
        """
        String representation of the Mocked instance.

        Includes configuration details and formatted history if available.

        :return: A string representation showing disposable, prefetch_history settings,
                 and optionally the formatted history.
        """
        if self._history_formatter:
            formatted = None
            if self._history is not None:
                formatted = self._history_formatter.format_history(self._history)
            return (f"Mocked<disposable={self._disposable}, "
                    f"prefetch_history={self._prefetch_history} "
                    f"history={formatted}>")
        else:
            return (f"Mocked<disposable={self._disposable}, "
                    f"prefetch_history={self._prefetch_history}>")



from typing import Optional, Union

from jj.expiration_policy import ExpirationPolicy
from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import HistoryAdapterType, default_history_adapter
from ._remote_handler import RemoteHandler
from ._remote_mock import RemoteMock
from ._remote_response import RemoteResponseType
from ._settings import get_remote_mock_url

__all__ = ("create_remote_handler",)


def create_remote_handler(matcher: Union[RequestMatcher, LogicalMatcher],
                          response: RemoteResponseType,
                          expiration_policy: Optional[ExpirationPolicy] = None,
                          *,
                          mock_url: Optional[str] = None,
                          history_adapter: Optional[HistoryAdapterType] = default_history_adapter
                          ) -> RemoteHandler:
    """
    Create a remote handler for mocking HTTP requests.

    This function creates a `RemoteHandler` instance that will intercept and mock
    HTTP requests matching the specified criteria, returning the configured response.

    :param matcher: A matcher that determines which requests to intercept. Can be either
                    a `RequestMatcher` or `LogicalMatcher`.
    :param response: The response to return for matched requests. Can be any valid
                     remote response type.
    :param expiration_policy: Optional policy that determines when the mock should expire.
                              If `None`, the mock will not expire.
    :param mock_url: Optional URL of the remote mock server. If `None`, uses the value
                     from the `JJ_REMOTE_MOCK_URL` environment variable or defaults to
                     "http://localhost:8080".
    :param history_adapter: Optional adapter for handling request/response history.
                            Defaults to `default_history_adapter`.
    :return: A `RemoteHandler` instance configured with the specified parameters.
    """
    if mock_url is None:
        mock_url = get_remote_mock_url()
    return RemoteMock(mock_url).create_handler(matcher, response, expiration_policy,
                                               history_adapter=history_adapter)



from .._resolvable_matcher import ResolvableMatcher

__all__ = ("LogicalMatcher",)


class LogicalMatcher(ResolvableMatcher):
    """
    Serves as a base class for logical matchers that combine multiple matchers.

    Logical matchers, such as `AnyMatcher` and `AllMatcher`, use this class as
    a foundation to combine several `ResolvableMatcher` instances, enabling
    the creation of more complex matching rules.
    """
    pass



from typing import Any, Dict, Union

from aiohttp.web_urldispatcher import DynamicResource
from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("RouteMatcher",)


class _Resource(DynamicResource):
    """
    A resource that defines the logic for matching URL paths.

    This class extends `DynamicResource` from aiohttp and provides the
    ability to match a URL path against a predefined route pattern.
    """

    def match(self, path: str) -> Union[Dict[str, str], None]:
        """
        Match the given path against the route pattern.

        :param path: The URL path to match.
        :return: A dictionary of matched segments if the path matches, otherwise `None`.
        """
        return self._match(path)


@packable("jj.matchers.RouteMatcher")
class RouteMatcher(AttributeMatcher):
    """
    Matches HTTP requests based on a URL route pattern.

    This matcher evaluates whether the request path matches a predefined route pattern,
    and can extract segments from the path.
    """

    def __init__(self, path: str) -> None:
        """
        Initialize a RouteMatcher with the expected route path.

        :param path: The URL route pattern to match.
        """
        self._path = path
        self._resource = _Resource(path)

    @property
    def path(self) -> str:
        """
        Return the expected route path.

        :return: The route pattern.
        """
        return self._path

    def get_segments(self, path: str) -> Dict[str, str]:
        """
        Extract and return segments from the given path.

        :param path: The URL path to extract segments from.
        :return: A dictionary of matched segments.
        """
        return self._resource.match(path) or {}

    async def match(self, path: str) -> bool:
        """
        Determine if the actual path matches the expected route pattern.

        :param path: The URL path to match.
        :return: `True` if the path matches the route pattern, otherwise `False`.
        """
        return self._resource.match(path) is not None

    def __repr__(self) -> str:
        """
        Return a string representation of the RouteMatcher instance.

        :return: A string describing the class and expected route pattern.
        """
        return f"{self.__class__.__qualname__}({self._path!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the RouteMatcher instance for serialization.

        :return: A dictionary containing the expected route path.
        """
        return {"path": self._path}

    @classmethod
    def __unpacked__(cls, *, path: str, **kwargs: Any) -> "RouteMatcher":
        """
        Unpack a RouteMatcher instance from its serialized form.

        :param path: The expected route pattern to use for this matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of RouteMatcher.
        """
        return cls(path)



import json
from datetime import datetime
from typing import Any, List, Optional, TypedDict, Union, cast

from jj import DelayedResponse, RelayResponse, Response
from jj.expiration_policy import ExpirationPolicy
from jj.matchers import LogicalMatcher, RequestMatcher
from jj.mock import HistoryItem

from ._history import BodyParser

__all__ = ("JsonRenderer", "Handler",)


class Handler(TypedDict):
    id: str
    expiration_policy: Union[ExpirationPolicy, None]
    matcher: Union[RequestMatcher, LogicalMatcher]
    response: Union[Response, RelayResponse, DelayedResponse]
    registered_at: datetime
    history_url: str


class JsonRenderer:
    def __init__(self, body_parser: Optional[BodyParser] = None) -> None:
        self._body_parser = body_parser or BodyParser()

    def render_handlers(self, handlers: List[Handler]) -> str:
        result = []
        for handler in handlers:
            result.append({
                "id": handler["id"],
                "registered_at": self._pack(handler["registered_at"]),
                "expiration_policy": self._pack(handler["expiration_policy"]),
                "matcher": self._pack(handler["matcher"]),
                "response": self._pack_response(handler["response"]),
                "history": {
                    "url": handler["history_url"],
                },
            })
        return self._to_json(result)

    def render_history(self, history_items: List[HistoryItem]) -> str:
        result = []
        for item in history_items:
            item = self._body_parser.parse(item)
            result.append({
                "request": self._pack(item["request"]),
                "response": self._pack(item["response"]),
                "created_at": self._pack(item["created_at"]),
            })
        return self._to_json(result)

    def _pack_response(self, response: Union[Response, RelayResponse, DelayedResponse]) -> Any:
        packed = self._pack(response)
        if isinstance(response, (Response, DelayedResponse)):
            cls_name = response.__class__.__name__
            body = self._body_parser._parse_by_content_type(response.content_type,
                                                            cast(bytes, response.body))
            packed[cls_name]["body"] = body
        return packed

    def _pack(self, value: Any) -> Any:
        if hasattr(value, "__packed__"):
            return {value.__class__.__name__: self._pack(value.__packed__())}
        elif isinstance(value, dict):
            return {k: self._pack(v) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [self._pack(v) for v in value]
        elif isinstance(value, datetime):
            return value.isoformat()
        else:
            return value

    def _to_json(self, value: Any) -> str:
        return json.dumps(value, default=str, ensure_ascii=False, indent=4)



from datetime import datetime
from typing import Any, Callable, Dict, Optional, Tuple, Union

from packed import pack, unpack
from yarl import URL

import jj
from jj import default_app, default_handler
from jj.apps import BaseApp, create_app
from jj.expiration_policy import ExpirationPolicy
from jj.http.codes import BAD_REQUEST, OK
from jj.http.headers import CONTENT_TYPE
from jj.http.methods import ANY, DELETE, GET, POST
from jj.matchers import LogicalMatcher, RequestMatcher, ResolvableMatcher, exists
from jj.requests import Request
from jj.resolvers import Registry, Resolver
from jj.responses import Response, StreamResponse, TemplateResponse

from ..handlers import HandlerFunction
from ._history import HistoryRepository
from ._json_renderer import JsonRenderer
from ._remote_response import REMOTE_RESPONSES, RemoteResponseType

__all__ = ("Mock",)


MatcherType = Union[RequestMatcher, LogicalMatcher]


class _DecodeError(Exception):
    pass


class Mock(jj.App):
    def __init__(self,
                 app_factory: Callable[..., BaseApp] = create_app,
                 resolver_factory: Callable[..., Resolver] = Resolver) -> None:
        self._resolver = resolver_factory(Registry(), default_app, default_handler)
        self._app = app_factory(resolver=self._resolver)
        self._repo = HistoryRepository()
        self._renderer = JsonRenderer()

    def _decode(self, payload: bytes) -> Tuple[str, MatcherType, RemoteResponseType,
                                               Optional[ExpirationPolicy]]:
        def resolver(cls: Any, **kwargs: Any) -> Any:
            return cls.__unpacked__(**kwargs, resolver=self._resolver)

        try:
            decoded = unpack(payload, {ResolvableMatcher: resolver})
        except Exception as e:
            raise _DecodeError(f"Decode Error: can't unpack message ({e})")

        errors = []

        handler_id = decoded.get("id")
        if not isinstance(handler_id, str):
            errors.append(f"Decode Error: invalid handler id ({handler_id!r})")

        matcher = decoded.get("request")
        if not isinstance(matcher, (RequestMatcher, LogicalMatcher)):
            errors.append(f"Decode Error: invalid request field ({matcher!r})")

        response = decoded.get("response")
        if not isinstance(response, REMOTE_RESPONSES):
            errors.append(f"Decode Error: invalid response field ({response!r})")

        expiration_policy = decoded.get("expiration_policy")
        if not isinstance(expiration_policy, (ExpirationPolicy, type(None))):
            errors.append(f"Decode Error: invalid expiration policy ({expiration_policy!r})")

        if len(errors) > 0:
            raise _DecodeError("\n".join(errors))

        return handler_id, matcher, response, expiration_policy

    def _register_handler(self, handler_id: str,
                          matcher: MatcherType,
                          response: RemoteResponseType,
                          expiration_policy: Optional[ExpirationPolicy]) -> None:
        async def handler(req: Request) -> RemoteResponseType:
            if isinstance(response, TemplateResponse):
                # Required to populate request body for template rendering
                await req.read()
                await req.post()
            res = response.copy()
            await res._prepare_hook(req)
            return res

        self._save_handler_info(handler, handler_id, matcher, response, expiration_policy)

        setattr(self._app.__class__, handler_id, matcher(handler))

    def _deregister_handler(self, handler_id: str) -> None:
        handler = getattr(self._app.__class__, handler_id, None)
        if handler is None:
            return

        try:
            delattr(self._app.__class__, handler_id)
        except AttributeError:
            pass

        matchers = self._resolver.get_matchers(handler)
        for matcher in matchers:
            self._resolver.deregister_matcher(matcher, handler)

        attributes = self._resolver.get_attributes(handler)
        for attribute in attributes:
            self._resolver.deregister_attribute(attribute, handler)

        self._resolver._registry.remove_container(handler)

    @jj.match(POST, "/__jj__/reset", headers={"x-jj-remote-mock": exists})
    async def reset(self, request: Request) -> Response:
        await request.read()

        handlers = self._resolver.get_handlers(self._app.__class__)
        for handler in handlers:
            handler_id = self._resolver.get_attribute("handler_id", handler, None)
            if handler_id:
                self._deregister_handler(handler_id)

        await self._repo.clear()

        return Response(status=OK, json={"status": OK})

    @jj.match_any([
        jj.match(POST, "/__jj__/register", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(POST, headers={"x-jj-remote-mock": exists})
    ])
    async def register(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, matcher, response, expiration_policy = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        self._register_handler(handler_id, matcher, response, expiration_policy)

        return Response(status=OK, json={"status": OK})

    @jj.match_any([
        jj.match(DELETE, "/__jj__/deregister", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(DELETE, headers={"x-jj-remote-mock": exists})
    ])
    async def deregister(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, *_ = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        self._deregister_handler(handler_id)
        await self._repo.delete_by_tag(handler_id)  # delete history

        return Response(status=OK, json={"status": OK})

    @jj.match_any([
        jj.match(GET, "/__jj__/history", headers={"x-jj-remote-mock": exists}),
        # backward compatibility
        jj.match(GET, headers={"x-jj-remote-mock": exists})
    ])
    async def history(self, request: Request) -> Response:
        payload = await request.read()
        try:
            handler_id, *_ = self._decode(payload)
        except Exception as e:
            return Response(status=BAD_REQUEST, json={"status": BAD_REQUEST, "error": str(e)})

        history = await self._repo.get_by_tag(handler_id)
        packed = pack(history)
        return Response(status=OK, body=packed)

    def _get_base_url(self, request_url: URL) -> str:
        base_url = f"{request_url.scheme}://{request_url.host}"
        if request_url.port not in (80, 443):
            base_url += f":{request_url.port}"
        return base_url

    def _save_handler_info(self, handler: HandlerFunction,
                           handler_id: str,
                           matcher: MatcherType,
                           response: RemoteResponseType,
                           expiration_policy: Optional[ExpirationPolicy]) -> None:
        self._resolver.register_attribute("handler_id", handler_id, handler)
        self._resolver.register_attribute("expiration_policy", expiration_policy, handler)
        self._resolver.register_attribute("matcher", matcher, handler)
        self._resolver.register_attribute("response", response, handler)
        self._resolver.register_attribute("registered_at", datetime.utcnow(), handler)

    def _get_handler_info(self, handler: HandlerFunction) -> Dict[str, Any]:
        return {
            "id": self._resolver.get_attribute("handler_id", handler),
            "expiration_policy": self._resolver.get_attribute("expiration_policy", handler),
            "matcher": self._resolver.get_attribute("matcher", handler),
            "response": self._resolver.get_attribute("response", handler),
            "registered_at": self._resolver.get_attribute("registered_at", handler),
        }

    def _get_handler_by_id(self, handler_id: str) -> Union[HandlerFunction, None]:
        for handler in self._resolver.get_handlers(self._app.__class__):
            if self._resolver.get_attribute("handler_id", handler) == handler_id:
                return handler
        return None

    @jj.match_any([
        jj.match(GET, "/__jj__"),
        jj.match(GET, "/__jj__/")
    ])
    async def api_index(self, request: Request) -> Response:
        base_url = self._get_base_url(request.url)
        handlers_url = f"{base_url}/__jj__/handlers"
        return jj.Response(status=OK, json={"handlers": {"url": handlers_url}})

    @jj.match(GET, "/__jj__/handlers")
    async def api_handlers(self, request: Request) -> Response:
        base_url = self._get_base_url(request.url)
        handler_list = self._resolver.get_handlers(self._app.__class__)
        handlers = []
        for handler in reversed(handler_list):
            handler_info = self._get_handler_info(handler)
            handler_info.update({
                "history_url": f"{base_url}/__jj__/handlers/{handler_info['id']}/history"
            })
            handlers.append(handler_info)
        body = self._renderer.render_handlers(handlers)  # type: ignore
        return Response(status=OK, body=body, headers={CONTENT_TYPE: "application/json"})

    @jj.match(GET, "/__jj__/handlers/{handler_id}/history")
    async def api_history(self, request: Request) -> Response:
        handler_id = request.segments["handler_id"]
        if not self._get_handler_by_id(handler_id):
            return Response(status=BAD_REQUEST,
                            json={"status": BAD_REQUEST, "error": "Handler not found"})
        history = await self._repo.get_by_tag(handler_id)
        body = self._renderer.render_history(history)
        return Response(status=OK, body=body, headers={CONTENT_TYPE: "application/json"})

    @jj.match(ANY)
    async def resolve(self, request: Request) -> StreamResponse:
        handler = await self._resolver.resolve(request, self._app)
        response = await handler(request)

        handler_id = self._resolver.get_attribute("handler_id", handler, default=None)
        if handler_id:
            await self._repo.add(request, response, tags=[handler_id])

        return response



from inspect import isclass
from typing import Any, List, Optional, Type, Union
from unittest.mock import sentinel as nil

from undecorated import undecorated

from jj.expiration_policy import ExpirationPolicy

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ..requests import Request
from ._matcher_function import MatcherFunction
from ._registry import Registry

__all__ = ("Resolver",)

AppOrHandler = Union[Type[AbstractApp], HandlerFunction]


class Resolver:
    def __init__(self, registry: Registry,
                 default_app: AbstractApp,
                 default_handler: HandlerFunction) -> None:
        self._registry = registry
        self._default_app = default_app
        self._default_handler = default_handler

    def unwrap(self, fn: Any) -> Any:
        try:
            unwrapped = undecorated(fn)
        except ValueError:
            return fn
        return fn if (unwrapped is None) else unwrapped

    # Apps

    def register_app(self, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.add(None, "apps", app)

    def deregister_app(self, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.remove(None, "apps", app)

    def get_apps(self) -> List[Type[AbstractApp]]:
        apps = self._registry.get(None, "apps")
        return list(apps.keys())

    # Handlers

    def register_handler(self, handler: HandlerFunction, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self.deregister_handler(handler, type(self._default_app))
        self.register_app(app)
        self._registry.add(app, "handlers", handler)

    def deregister_handler(self, handler: HandlerFunction, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.remove(app, "handlers", handler)

    def get_handlers(self, app: Type[AbstractApp]) -> List[HandlerFunction]:
        assert isclass(app)
        handlers = self._registry.get(app, "handlers")
        return list(handlers.keys())

    # Matchers

    def register_matcher(self, matcher: MatcherFunction, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self.register_handler(unwrapped, type(self._default_app))
        self._registry.add(unwrapped, "matchers", matcher)

    def deregister_matcher(self, matcher: MatcherFunction, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.remove(unwrapped, "matchers", matcher)

        matchers = self._registry.get(unwrapped, "matchers")
        if len(matchers) == 0:
            self._registry.remove_name(unwrapped, "matchers")

    def get_matchers(self, handler: HandlerFunction) -> List[MatcherFunction]:
        unwrapped = self.unwrap(handler)
        matchers = self._registry.get(unwrapped, "matchers")
        return list(matchers.keys())

    # Attributes

    def register_attribute(self, name: Any, value: Any, handler: AppOrHandler) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.add(unwrapped, "attributes", name, value)

    def deregister_attribute(self, attribute_name: Any, handler: AppOrHandler) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.remove(unwrapped, "attributes", attribute_name)

        attributes = self._registry.get(unwrapped, "attributes")
        if len(attributes) == 0:
            self._registry.remove_name(unwrapped, "attributes")

    def get_attribute(self, attribute_name: Any,
                      handler: AppOrHandler,
                      default: Any = nil) -> Any:
        unwrapped = self.unwrap(handler)
        attributes = self._registry.get(unwrapped, "attributes")
        return attributes.get(attribute_name, default)

    def get_attributes(self, handler: AppOrHandler) -> List[Any]:
        unwrapped = self.unwrap(handler)
        attributes = self._registry.get(unwrapped, "attributes")
        return list(attributes.keys())

    # Resolve

    async def _match_request(self, request: Request, matchers: List[MatcherFunction]) -> bool:
        if len(matchers) == 0:
            return False
        for matcher in matchers:
            if not await matcher(request):
                return False
        return True

    # ExpirationPolicy

    async def _is_handler_expired(self, handler: HandlerFunction, request: Request) -> bool:
        expiration_policy: Optional[ExpirationPolicy] = self.get_attribute(
            "expiration_policy", handler, default=None
        )
        if expiration_policy is None:
            return False

        return await expiration_policy.is_expired(request)

    async def resolve(self, request: Request, app: AbstractApp) -> HandlerFunction:
        assert not isclass(app)
        handlers = self.get_handlers(type(app))
        for handler in reversed(handlers):
            matchers = self.get_matchers(handler)
            if await self._match_request(request, matchers):
                if await self._is_handler_expired(handler, request):
                    continue
                return handler
        return self._default_handler



from multidict import istr

ACCEPT = istr("Accept")
ACCEPT_CHARSET = istr("Accept-Charset")
ACCEPT_ENCODING = istr("Accept-Encoding")
ACCEPT_LANGUAGE = istr("Accept-Language")
ACCEPT_RANGES = istr("Accept-Ranges")
ACCESS_CONTROL_ALLOW_CREDENTIALS = istr("Access-Control-Allow-Credentials")
ACCESS_CONTROL_ALLOW_HEADERS = istr("Access-Control-Allow-Headers")
ACCESS_CONTROL_ALLOW_METHODS = istr("Access-Control-Allow-Methods")
ACCESS_CONTROL_ALLOW_ORIGIN = istr("Access-Control-Allow-Origin")
ACCESS_CONTROL_EXPOSE_HEADERS = istr("Access-Control-Expose-Headers")
ACCESS_CONTROL_MAX_AGE = istr("Access-Control-Max-Age")
ACCESS_CONTROL_REQUEST_HEADERS = istr("Access-Control-Request-Headers")
ACCESS_CONTROL_REQUEST_METHOD = istr("Access-Control-Request-Method")
AGE = istr("Age")
ALLOW = istr("Allow")
ALT_SVC = istr("Alt-Svc")
AUTHORIZATION = istr("Authorization")
CACHE_CONTROL = istr("Cache-Control")
CLEAR_SITE_DATA = istr("Clear-Site-Data")
CONNECTION = istr("Connection")
CONTENT_DISPOSITION = istr("Content-Disposition")
CONTENT_ENCODING = istr("Content-Encoding")
CONTENT_LANGUAGE = istr("Content-Language")
CONTENT_LENGTH = istr("Content-Length")
CONTENT_LOCATION = istr("Content-Location")
CONTENT_RANGE = istr("Content-Range")
CONTENT_SECURITY_POLICY = istr("Content-Security-Policy")
CONTENT_SECURITY_POLICY_REPORT_ONLY = istr("Content-Security-Policy-Report-Only")
CONTENT_TYPE = istr("Content-Type")
COOKIE = istr("Cookie")
COOKIE2 = istr("Cookie2")
DNT = istr("DNT")
DATE = istr("Date")
ETAG = istr("ETag")
EARLY_DATA = istr("Early-Data")
EXPECT = istr("Expect")
EXPECT_CT = istr("Expect-CT")
EXPIRES = istr("Expires")
FEATURE_POLICY = istr("Feature-Policy")
FORWARDED = istr("Forwarded")
FROM = istr("From")
HOST = istr("Host")
IF_MATCH = istr("If-Match")
IF_MODIFIED_SINCE = istr("If-Modified-Since")
IF_NONE_MATCH = istr("If-None-Match")
IF_RANGE = istr("If-Range")
IF_UNMODIFIED_SINCE = istr("If-Unmodified-Since")
INDEX = istr("Index")
KEEP_ALIVE = istr("Keep-Alive")
LARGE_ALLOCATION = istr("Large-Allocation")
LAST_MODIFIED = istr("Last-Modified")
LOCATION = istr("Location")
ORIGIN = istr("Origin")
PRAGMA = istr("Pragma")
PROXY_AUTHENTICATE = istr("Proxy-Authenticate")
PROXY_AUTHORIZATION = istr("Proxy-Authorization")
PUBLIC_KEY_PINS = istr("Public-Key-Pins")
PUBLIC_KEY_PINS_REPORT_ONLY = istr("Public-Key-Pins-Report-Only")
RANGE = istr("Range")
REFERER = istr("Referer")
REFERRER_POLICY = istr("Referrer-Policy")
RETRY_AFTER = istr("Retry-After")
SEC_WEBSOCKET_ACCEPT = istr("Sec-WebSocket-Accept")
SERVER = istr("Server")
SERVER_TIMING = istr("Server-Timing")
SET_COOKIE = istr("Set-Cookie")
SET_COOKIE2 = istr("Set-Cookie2")
SOURCEMAP = istr("SourceMap")
STRICT_TRANSPORT_SECURITY = istr("Strict-Transport-Security")
TE = istr("TE")
TIMING_ALLOW_ORIGIN = istr("Timing-Allow-Origin")
TK = istr("Tk")
TRAILER = istr("Trailer")
TRANSFER_ENCODING = istr("Transfer-Encoding")
UPGRADE_INSECURE_REQUESTS = istr("Upgrade-Insecure-Requests")
USER_AGENT = istr("User-Agent")
VARY = istr("Vary")
VIA = istr("Via")
WWW_AUTHENTICATE = istr("WWW-Authenticate")
WARNING = istr("Warning")
X_CONTENT_TYPE_OPTIONS = istr("X-Content-Type-Options")
X_DNS_PREFETCH_CONTROL = istr("X-DNS-Prefetch-Control")
X_FORWARDED_FOR = istr("X-Forwarded-For")
X_FORWARDED_HOST = istr("X-Forwarded-Host")
X_FORWARDED_PROTO = istr("X-Forwarded-Proto")
X_FRAME_OPTIONS = istr("X-Frame-Options")
X_XSS_PROTECTION = istr("X-XSS-Protection")



import logging
from logging import LogRecord

__all__ = ("RequestFilter",)


class RequestFilter(logging.Filter):
    def filter(self, record: LogRecord) -> bool:
        request = getattr(record, "jj_request", None)
        response = getattr(record, "jj_response", None)
        if (request is not None) and (response is None):
            return False
        return super().filter(record) is not False



import logging

__all__ = ("Logger",)


class Logger(logging.Logger):
    def clearHandlers(self) -> "Logger":
        for handler in self.handlers:
            self.removeHandler(handler)
        return self



from typing import Any, Dict

from packed import packable

from ..requests import Request
from ._expiration_policy import ExpirationPolicy

__all__ = ("ExpireAfterRequests",)


@packable("jj.expiration_policy.ExpireAfterRequests")
class ExpireAfterRequests(ExpirationPolicy):
    """
    Represents an expiration policy based on the number of requests.

    This policy expires a mocked response after a specified number of requests.
    Once the maximum request count is reached, the response is considered expired
    and will no longer match subsequent requests.
    """

    def __init__(self, max_requests_count: int) -> None:
        """
        Initialize the ExpireAfterRequests policy with a maximum number of allowed requests.

        :param max_requests_count: The maximum number of requests allowed before the mock expires.
                                   Must be greater than 0.
        :raises AssertionError: If `max_requests_count` is less than or equal to 0.
        """
        assert max_requests_count > 0, \
            f'max_requests_count must be more than 0, {max_requests_count} given'

        self._max_requests_count = max_requests_count
        self._current_requests_count = 0

    @property
    def max_requests_count(self) -> int:
        """
        Return the maximum number of requests allowed before expiration.

        :return: The maximum number of requests.
        """
        return self._max_requests_count

    async def is_expired(self, request: Request) -> bool:
        """
        Check if the mock has expired based on the number of requests made.

        :param request: The current request being checked against the expiration policy.
        :return: `True` if the mock has expired, otherwise `False`.
        """
        if self._current_requests_count < self._max_requests_count:
            self._current_requests_count += 1
            return False
        return True

    def __repr__(self) -> str:
        """
        Return a string representation of the ExpireAfterRequests instance.

        :return: A string representation of the instance.
        """
        return f"{self.__class__.__qualname__}({self._max_requests_count!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Serialize the ExpireAfterRequests instance for packing.

        :return: A dictionary containing the serialized form of the instance.
        """
        return {
            "max_requests_count": self._max_requests_count,
        }

    @classmethod
    def __unpacked__(cls, *, max_requests_count: int, **kwargs: Any) -> "ExpireAfterRequests":
        """
        Unpack a serialized ExpireAfterRequests instance.

        :param max_requests_count: The maximum number of requests allowed before expiration.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of ExpireAfterRequests.
        """
        return cls(max_requests_count=max_requests_count)



import io
from http.cookies import Morsel
from json import dumps
from typing import Any, Dict, List, Optional, Tuple, Union
from unittest.mock import sentinel as nil

from aiohttp import web
from aiohttp.abc import AbstractStreamWriter
from aiohttp.typedefs import LooseHeaders
from aiohttp.web import ContentCoding
from aiohttp.web_request import BaseRequest
from multidict import CIMultiDict
from packed import packable

from ..http.headers import CONTENT_DISPOSITION, CONTENT_TYPE
from ._stream_response import StreamResponse
from ._utils import cookie_to_dict, get_response_body

__all__ = ("Response",)


@packable("jj.responses.Response")
class Response(web.Response, StreamResponse):
    """
    Represents an HTTP response that can handle various content types, including
    JSON, plain text, and file streams, while supporting chunked encoding,
    compression, and cookie management.

    This class extends aiohttp's `web.Response` and adds additional functionality
    for handling response bodies, cookies, headers, and content encoding. It also
    supports serialization and deserialization for remote handling of responses.
    """

    def __init__(self, *,
                 json: Any = nil,
                 body: Optional[Union[str, bytes]] = None,
                 text: Optional[str] = None,
                 status: int = 200,
                 reason: Optional[str] = None,
                 headers: Optional[LooseHeaders] = None) -> None:
        """
        Initialize the Response object.

        :param json: The JSON object to be used as the response body.
                     If provided, overrides `body` and `text`.
        :param body: The raw body content of the response as a string or bytes.
        :param text: The plain text content of the response.
        :param status: The HTTP status code for the response (default is 200).
        :param reason: The reason phrase for the response status code.
        :param headers: A dictionary or list of headers to include in the response.
        """
        headers = CIMultiDict(headers or {})

        if json is not nil:
            assert (body is None) and (text is None)
            body = dumps(json)
            if CONTENT_TYPE not in headers:
                headers[CONTENT_TYPE] = "application/json"

        if (body is None) and (text is None):
            body = ""

        if isinstance(body, io.IOBase):
            if CONTENT_DISPOSITION not in headers:
                headers[CONTENT_DISPOSITION] = "inline"

        super().__init__(body=body, text=text, status=status, reason=reason, headers=headers)

    @property
    def content_coding(self) -> Optional[ContentCoding]:
        """
        Get the content encoding for the response.

        :return: The content encoding if compression is enabled, otherwise None.
        """
        return self._compression_force

    def _cookie_to_dict(self, cookie: "Morsel[str]") -> Dict[str, Union[str, None]]:
        """
        Convert a Morsel cookie object to a dictionary.

        :param cookie: The Morsel object representing the cookie.
        :return: A dictionary representing the cookie's attributes.
        """
        return cookie_to_dict(cookie)

    def copy(self) -> "Response":
        """
        Create a copy of the Response object.

        This method clones the current response including its status, reason, headers,
        cookies, and compression settings.

        :return: A new instance of the Response object.
        :raises AssertionError: If the response is already prepared.
        """
        assert not self.prepared

        response = self.__class__(status=self.status, reason=self.reason,
                                  headers=self.headers, body=self.body)  # type: ignore
        for cookie in self.cookies.values():
            response.set_cookie(**cookie_to_dict(cookie))  # type: ignore
        if self.chunked:
            response.enable_chunked_encoding()
        if self._compression_force:
            response.enable_compression(self._compression_force)
        return response

    def get_body(self) -> bytes:
        """
        Get the body of the response.

        This method provides backward compatibility for retrieving the response body
        as bytes.

        :return: The response body as bytes.
        """
        return get_response_body(self.body)

    async def _prepare_hook(self, request: BaseRequest) -> "Response":
        """
        Hook method to be executed before preparing the response.

        This method allows customization before the response is sent to the client.

        :param request: The request object associated with the response.
        :return: The Response object itself.
        """
        return self

    async def prepare(self, request: BaseRequest) -> Optional[AbstractStreamWriter]:
        """
        Prepare the response for sending to the client.

        This method finalizes the response, making it ready to be transmitted.

        :param request: The incoming HTTP request.
        :return: The stream writer used to send the response.
        """
        await self._prepare_hook(request)
        return await super().prepare(request)

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the Response object into a dictionary for serialization.

        This method is used for converting the Response object into a serializable
        format that can be transmitted or stored remotely.

        :return: A dictionary containing the packed response data.
        :raises AssertionError: If the response is already prepared.
        """
        assert not self.prepared

        headers = [[key, val] for key, val in self.headers.items()]
        cookies = [cookie_to_dict(cookie) for cookie in self.cookies.values()]

        body = get_response_body(self.body)

        compression: Optional[str] = None
        if isinstance(self._compression_force, ContentCoding):
            compression = self._compression_force.value
        else:
            compression = self._compression_force

        return {
            "status": self.status,
            "reason": self.reason,
            "headers": headers,
            "cookies": cookies,
            "body": body,
            "chunked": self.chunked,
            "compression": compression,
        }

    @classmethod
    def __unpacked__(cls, *,
                     status: int,
                     reason: Optional[str],
                     headers: List[Tuple[str, str]],
                     cookies: List[Dict[str, Union[str, None]]],
                     body: Optional[bytes],
                     chunked: bool,
                     compression: Optional[ContentCoding],
                     **kwargs: Any) -> "Response":
        """
        Reconstruct a Response instance from unpacked parameters.

        :param status: The HTTP status code.
        :param reason: The reason phrase for the HTTP status.
        :param headers: A list of headers for the response.
        :param cookies: A list of cookie dictionaries.
        :param body: The response body as bytes.
        :param chunked: A flag indicating whether chunked encoding is enabled.
        :param compression: The content compression coding.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new Response object.
        """
        response = cls(status=status, reason=reason, headers=headers, body=body)
        for cookie in cookies:
            response.set_cookie(**cookie)  # type: ignore
        if compression:
            response.enable_compression(ContentCoding(compression))
        if chunked:
            response.enable_chunked_encoding()
        return response



from ..attribute_matchers import StrOrAttrMatcher
from ._header_matcher import HeaderMatcher
from ._method_matcher import MethodMatcher
from ._param_matcher import DictOrTupleListOrAttrMatcher, ParamMatcher
from ._path_matcher import PathMatcher
from ._request_matcher import RequestMatcher

__all__ = (
    "DictOrTupleListOrAttrMatcher",
    "HeaderMatcher",
    "MethodMatcher",
    "ParamMatcher",
    "PathMatcher",
    "RequestMatcher",
    "StrOrAttrMatcher",
)



import importlib.util
from importlib.abc import Loader
from pathlib import Path
from types import ModuleType

__all__ = ("load_module",)

from typing import cast


def _path_to_module_name(path: Path) -> str:
    """
    Convert a file path to a valid Python module name.

    This helper function transforms the given file path into a module name by removing
    the file extension and converting the path structure into a dot-separated string,
    as required for module names. If the path is absolute, the leading part of the
    path is removed.

    :param path: The file path to be converted.
    :return: A string representing the module name derived from the path.
    """
    parts = path.with_suffix("").parts
    if path.is_absolute():
        parts = parts[1:]
    return ".".join(parts)


def load_module(path: Path) -> ModuleType:
    """
    Load a Python module from the specified file path.

    This function dynamically loads a module from the given path, using the
    importlib utilities to create and load the module. It converts the file path
    into a valid module name and registers the module in `sys.modules`.

    :param path: The file path from which to load the module.
    :return: The loaded module as a `ModuleType` object.
    :raises ModuleNotFoundError: If the module cannot be found or loaded.
    """
    module_name = _path_to_module_name(path)
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise ModuleNotFoundError(module_name)

    module = importlib.util.module_from_spec(spec)
    cast(Loader, spec.loader).exec_module(module)

    return module



from collections import OrderedDict
from typing import Any, MutableMapping, Type

__all__ = ("Registry",)


class Registry:
    def __init__(self,
                 mutable_mapping_factory: Type[MutableMapping[Any, Any]] = OrderedDict) -> None:
        self._factory = mutable_mapping_factory
        self._registry = self._factory()

    def add(self, container: Any, name: str, key: Any, value: Any = None) -> None:
        if container not in self._registry:
            self._registry[container] = self._factory()
        if name not in self._registry[container]:
            self._registry[container][name] = self._factory()
        self._registry[container][name][key] = value

    def get(self, container: Any, name: str) -> Any:
        if (container not in self._registry) or (name not in self._registry[container]):
            return self._factory()
        return self._registry[container][name]

    def remove(self, container: Any, name: str, key: Any) -> None:
        # backward compatibility
        return self.remove_key(container, name, key)

    def remove_key(self, container: Any, name: str, key: Any) -> None:
        if (container in self._registry) and (name in self._registry[container]):
            self._registry[container][name].pop(key, None)

    def remove_name(self, container: Any, name: str) -> None:
        if container in self._registry:
            self._registry[container].pop(name, None)

    def remove_container(self, container: Any) -> None:
        self._registry.pop(container, None)



import re
from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("RegexMatcher",)


@packable("jj.matchers.RegexMatcher")
class RegexMatcher(AttributeMatcher):
    """
    Matches an HTTP request based on a regular expression pattern.

    This matcher evaluates if the actual value (e.g., a URL path or string) matches
    the specified regular expression pattern.
    """

    def __init__(self, pattern: str, flags: int = 0) -> None:
        """
        Initialize a RegexMatcher with a regex pattern and optional flags.

        :param pattern: The regular expression pattern to match.
        :param flags: Optional regex flags (e.g., `re.IGNORECASE`).
        """
        self._pattern = pattern
        self._flags = flags
        self._compiled = re.compile(self._pattern, self._flags)

    @property
    def pattern(self) -> str:
        """
        Return the regular expression pattern.

        :return: The regex pattern.
        """
        return self._pattern

    @property
    def flags(self) -> int:
        """
        Return the regex flags.

        :return: The regex flags.
        """
        return self._flags

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value matches the regular expression pattern.

        :param actual: The value to match against the regex.
        :return: `True` if the value matches the pattern, otherwise `False`.
        """
        return self._compiled.search(actual) is not None

    def __repr__(self) -> str:
        """
        Return a string representation of the RegexMatcher instance.

        :return: A string describing the class, pattern, and flags.
        """
        if self._flags == 0:
            return f"{self.__class__.__qualname__}({self._pattern!r})"
        return f"{self.__class__.__qualname__}({self._pattern!r}, {self._flags!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the RegexMatcher instance for serialization.

        :return: A dictionary containing the pattern and flags.
        """
        return {"pattern": self._pattern, "flags": self._flags}

    @classmethod
    def __unpacked__(cls, *, pattern: str, flags: int = 0, **kwargs: Any) -> "RegexMatcher":
        """
        Unpack a RegexMatcher instance from its serialized form.

        :param pattern: The regex pattern to use for this matcher.
        :param flags: Optional regex flags.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of RegexMatcher.
        """
        return cls(pattern, flags)



import json
from typing import Any

from jj.http.headers import CONTENT_TYPE

from ._history_item import HistoryItem
from ._history_request import HistoryRequest
from ._history_response import HistoryResponse

__all__ = ("BodyParser",)


class BodyParser:
    def _parse_by_content_type(self, content_type: str, body: bytes) -> Any:
        if content_type.lower().startswith("application/json"):
            try:
                return json.loads(body)
            except:  # noqa: E722
                pass
        elif content_type.lower().startswith("text/plain"):
            try:
                return body.decode()
            except:  # noqa: E722
                pass
        return body

    def _parse_request_body(self, request: HistoryRequest) -> Any:
        content_type = request.headers.get(CONTENT_TYPE, "")
        return self._parse_by_content_type(content_type, request.body)

    def _parse_response_body(self, response: HistoryResponse) -> Any:
        content_type = response.headers.get(CONTENT_TYPE, "")
        return self._parse_by_content_type(content_type, response.body)

    def parse(self, history_item: HistoryItem) -> HistoryItem:
        request = history_item["request"].to_dict()
        request["body"] = self._parse_request_body(history_item["request"])
        history_item["request"] = history_item["request"].from_dict(request)

        response = history_item["response"].to_dict()
        response["body"] = self._parse_response_body(history_item["response"])
        history_item["response"] = history_item["response"].from_dict(response)

        return history_item



from typing import Any, Dict

from packed import packable

from ..requests import Request
from ._expiration_policy import ExpirationPolicy

__all__ = ("ExpireNever",)


@packable("jj.expiration_policy.ExpireNever")
class ExpireNever(ExpirationPolicy):
    """
    Represents a policy where a mocked response never expires.

    This policy ensures that the mocked response is always valid, regardless
    of how many requests have been made.
    """

    async def is_expired(self, request: Request) -> bool:
        """
        Check if the mock has expired. Always returns `False` as this policy never expires.

        :param request: The current request being evaluated (not used in this policy).
        :return: `False`, indicating the mock never expires.
        """
        return False

    def __repr__(self) -> str:
        """
        Return a string representation of the ExpireNever instance.

        :return: A string representation of the instance.
        """
        return f"{self.__class__.__qualname__}()"

    def __packed__(self) -> Dict[str, Any]:
        """
        Serialize the ExpireNever instance for packing.

        :return: A dictionary containing the serialized form of the instance.
        """
        return dict()

    @classmethod
    def __unpacked__(cls, **kwargs: Any) -> "ExpireNever":
        """
        Unpack a serialized ExpireNever instance.

        :return: A new instance of ExpireNever.
        """
        return cls()



from typing import Any, Dict

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, RouteMatcher, StrOrAttrMatcher
from ._request_matcher import RequestMatcher

__all__ = ("PathMatcher",)


@packable("jj.matchers.PathMatcher")
class PathMatcher(RequestMatcher):
    """
    Matches HTTP requests based on the URL path.

    This matcher checks if the incoming request's URL path matches a specific
    string or pattern, and can extract segments from the path if needed.
    """

    def __init__(self, path: StrOrAttrMatcher, *, resolver: Resolver) -> None:
        """
        Initialize a PathMatcher with the URL path and resolver.

        :param path: The path or pattern to match against, or a matcher that can
                     evaluate the path dynamically.
        :param resolver: The resolver responsible for registering this matcher.
        """
        super().__init__(resolver=resolver)
        if isinstance(path, AttributeMatcher):
            self._matcher = path
        else:
            self._matcher = RouteMatcher(path)

    @property
    def sub_matcher(self) -> AttributeMatcher:
        """
        Return the underlying attribute matcher used for the path matching.

        :return: The matcher that evaluates the path condition.
        """
        return self._matcher

    async def match(self, request: Request) -> bool:
        """
        Determine if the request path matches the expected path or pattern.

        :param request: The HTTP request containing the path to match.
        :return: `True` if the request path matches the expected path or pattern,
                 otherwise `False`.
        """
        matched = await self._matcher.match(request.path)
        if matched and isinstance(self._matcher, RouteMatcher):
            request.segments = self._matcher.get_segments(request.path)
        else:
            request.segments = None  # type: ignore
        return matched

    def __repr__(self) -> str:
        """
        Return a string representation of the PathMatcher instance.

        :return: A string describing the class, matcher, and resolver.
        """
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the PathMatcher instance for serialization.

        :return: A dictionary containing the serialized form of the matcher.
        """
        return {"path": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     path: StrOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "PathMatcher":
        """
        Unpack a PathMatcher instance from its serialized form.

        :param path: The path matcher to use for this instance.
        :param resolver: The resolver to bind this matcher to.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of PathMatcher.
        """
        return cls(path, resolver=resolver)



import codecs
from logging import LogRecord
from os import getenv
from string import Template

from multidict import MultiMapping

from ...requests import Request
from ...responses import Response
from ._formatter import Formatter

__all__ = ("TemplateFormatter",)

LOG_FORMAT = getenv("JJ_LOG_FORMAT", "-> $req_method $req_path\n<- $res_code $res_reason\n")
LOG_BODY_LIMIT = int(getenv("JJ_LOG_BODY_LIMIT", 10 ** 6))


class TemplateFormatter(Formatter):
    def __init__(self, log_format: str = LOG_FORMAT, log_body_limit: int = LOG_BODY_LIMIT) -> None:
        super().__init__()
        self._log_format = codecs.decode(log_format.encode(), 'unicode_escape')
        self._log_body_limit = log_body_limit
        self._template = Template(self._log_format)

    def format_response(self, response: Response, request: Request, record: LogRecord) -> str:
        try:
            return self._substitute(response, request)
        except KeyError as e:
            return f"Incorrect log format: ${e} does not exist"
        except Exception as e:
            return f"Error while formatting log: {repr(e)}"

    def _substitute(self, response: Response, request: Request) -> str:
        response_body = response.get_body()
        return self._template.substitute(
            # request
            req_method=request.method,
            req_path=request.url.path,
            req_query=request.url.query_string,
            req_headers=self._format_headers(request.headers),
            # response
            res_code=response.status,
            res_reason=response.reason,
            res_headers=self._format_headers(response.headers),
            res_body=response_body[:self._log_body_limit],
        )

    def _format_headers(self, headers: MultiMapping[str]) -> str:
        result = []
        for key, val in headers.items():
            result.append(f" {key}: {val}")
        return "\n".join(result)



from ..handlers import HandlerFunction
from ..requests import Request
from ..resolvers import Resolver

__all__ = ("ResolvableMatcher",)


class ResolvableMatcher:
    """
    Provides a base class for matchers that can resolve HTTP requests.

    This class defines the interface for a matcher that checks if an HTTP
    request meets certain conditions. It registers the matcher to a resolver
    and can be used to bind a handler function to matching requests.
    """

    def __init__(self, *, resolver: Resolver) -> None:
        """
        Initialize a ResolvableMatcher with a given resolver.

        :param resolver: The resolver responsible for handling the registration
                         of the matcher and its corresponding handler.
        """
        self._resolver = resolver

    async def match(self, request: Request) -> bool:
        """
        Determine if the incoming HTTP request matches the conditions of this matcher.

        :param request: The HTTP request to evaluate against the matcher.
        :return: `True` if the request matches, otherwise `False`.
        :raises NotImplementedError: This method must be implemented in a subclass.
        """
        raise NotImplementedError()

    def __call__(self, handler: HandlerFunction) -> HandlerFunction:
        """
        Register a handler function to be executed when a request matches.

        :param handler: The function that will handle requests matching this matcher.
        :return: The handler function, which is registered to the resolver.
        """
        self._resolver.register_matcher(self.match, handler)
        return handler

    def __repr__(self) -> str:
        """
        Return a string representation of the ResolvableMatcher instance.

        :return: A string describing the class and its resolver.
        """
        return f"{self.__class__.__qualname__}(resolver={self._resolver!r})"



from typing import Any, Optional, cast

from aiohttp import web
from aiohttp.typedefs import LooseHeaders
from multidict import CIMultiDict

from .._version import server_version

__all__ = ("StreamResponse",)


class StreamResponse(web.StreamResponse):
    """
    Represents an HTTP response that supports streaming.

    The `StreamResponse` class extends `aiohttp.web.StreamResponse` and adds a default
    `Server` header with the server version.

    :param status: The HTTP status code for the response (default is 200).
    :param reason: The reason phrase for the status code (optional).
    :param headers: Optional headers for the response, passed as a dictionary.
                    If no `Server` header is provided, it defaults to the server version.
    """

    def __init__(self, *,
                 status: int = 200,
                 reason: Optional[str] = None,
                 headers: Optional[LooseHeaders] = None,
                 **kwargs: Any) -> None:
        """
        Initialize the StreamResponse object with the specified status, reason, and headers.

        :param status: The HTTP status code for the response (default is 200).
        :param reason: The reason phrase for the status code (optional).
        :param headers: A dictionary of headers for the response (optional). If no `Server`
                        header is provided, it defaults to the current server version.
        """

        # _real_headers is an internal parameter used to pass a pre-populated
        # headers object. It is used by the `Response` class to avoid copying
        # the headers when creating a new response object. It is not intended
        # to be used by external code.
        _real_headers = kwargs.pop("_real_headers", None)

        if _real_headers is not None:
            headers = cast(CIMultiDict[str], _real_headers)
        else:
            headers = CIMultiDict(headers or {})
        headers["Server"] = headers.get("Server", server_version)

        super().__init__(status=status, reason=reason, headers=headers, **kwargs)



from typing import Any, Dict, List, Tuple, Union

from multidict import MultiDict, MultiMapping
from packed import packable

from ._attribute_matcher import AttributeMatcher
from ._equal_matcher import EqualMatcher
from ._exist_matcher import NotExistMatcher

__all__ = ("MultiDictMatcher", "StrOrAttrMatcher", "DictOrTupleList",)

StrOrAttrMatcher = Union[str, AttributeMatcher]
DictOrTupleList = Union[
    Dict[str, StrOrAttrMatcher],
    List[Tuple[str, StrOrAttrMatcher]],
]


@packable("jj.matchers.MultiDictMatcher")
class MultiDictMatcher(AttributeMatcher):
    """
    Matches HTTP requests based on multiple key-value pairs, such as query parameters or headers.

    This matcher checks if the actual request contains key-value pairs that match
    the expected key-value pairs, with the option to use either static values
    or dynamic matchers for each value.
    """

    def __init__(self, expected: DictOrTupleList) -> None:
        """
        Initialize a MultiDictMatcher with the expected key-value pairs.

        :param expected: The expected key-value pairs to match, provided as a dictionary
                         or list of tuples.
        """
        self._expected = MultiDict(expected)

    @property
    def expected(self) -> MultiDict[StrOrAttrMatcher]:
        """
        Return the expected key-value pairs for this matcher.

        :return: A MultiDict containing the expected key-value pairs.
        """
        return self._expected

    async def _match_any(self, submatcher: AttributeMatcher, values: List[Any]) -> bool:
        """
        Determine if any value in the list matches the given submatcher.

        :param submatcher: The matcher to evaluate each value.
        :param values: A list of values to be evaluated.
        :return: `True` if any value matches the submatcher, otherwise `False`.
        """
        for value in values:
            if await submatcher.match(value):
                return True
        return False

    async def match(self, actual: MultiMapping[str]) -> bool:
        """
        Determine if the actual request data satisfies the expected key-value conditions.

        :param actual: A MultiMapping representing request attributes (e.g., headers or
                       query parameters).
        :return: `True` if all expected conditions are met, otherwise `False`.
        """
        if not isinstance(actual, MultiMapping):
            raise TypeError(f"Expected MultiMapping, got '{type(actual).__name__}'")

        for key, val in self._expected.items():
            # 1) Special case: key must NOT exist
            if isinstance(val, NotExistMatcher):
                # If the key exists in `actual`, fail immediately
                if key in actual:
                    return False
                # Otherwise, key does not exist, so this condition is met
                continue

            # 2) Normal case: key must exist, and at least one value must match
            submatcher = val if isinstance(val, AttributeMatcher) else EqualMatcher(val)
            values: List[Any] = actual.getall(key, [])
            if not await self._match_any(submatcher, values):
                return False

        # All expected conditions are satisfied
        return True

    def __repr__(self) -> str:
        """
        Return a string representation of the MultiDictMatcher instance.

        :return: A string describing the class and expected key-value pairs.
        """
        expected = [(key, val) for key, val in self._expected.items()]
        return f"{self.__class__.__qualname__}({expected!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the MultiDictMatcher instance for serialization.

        :return: A dictionary containing the serialized expected key-value pairs.
        """
        expected = [[key, val] for key, val in self._expected.items()]
        return {"expected": expected}

    @classmethod
    def __unpacked__(cls, *, expected: DictOrTupleList, **kwargs: Any) -> "MultiDictMatcher":
        """
        Unpack a MultiDictMatcher instance from its serialized form.

        :param expected: The expected key-value pairs to use for this matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of MultiDictMatcher.
        """
        return cls(expected)



from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("ExistMatcher", "NotExistMatcher",)


@packable("jj.matchers.ExistMatcher")
class ExistMatcher(AttributeMatcher):
    """
    Matches any HTTP request.

    This matcher is used when the presence of a value is sufficient for matching,
    regardless of its actual content.
    """

    async def match(self, actual: Any) -> bool:
        """
        Always return `True`, as this matcher matches any request.

        :param actual: The value to match (unused in this case).
        :return: `True` for all requests.
        """
        return True

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the ExistMatcher instance for serialization.

        :return: An empty dictionary, as no specific state is stored.
        """
        return {}

    @classmethod
    def __unpacked__(cls, **kwargs: Any) -> "ExistMatcher":
        """
        Unpack an ExistMatcher instance from its serialized form.

        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of ExistMatcher.
        """
        return cls()


@packable("jj.matchers.NotExistMatcher")
class NotExistMatcher(AttributeMatcher):
    """
    Matches when an attribute does not exist in the HTTP request.

    This matcher is used to check the absence of a value in the request attributes.
    If the attribute exists, the matcher does not match.
    """

    async def match(self, actual: Any) -> bool:
        """
        Return `False` if the attribute exists in the request.

        This method is called when an attribute key is present. Since this matcher
        is designed to match when the key does not exist, it always returns `False`.

        :param actual: The value to check (not used, as existence is the only criteria).
        :return: `False`, indicating the key exists.
        """
        return False

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the NotExistMatcher instance for serialization.

        :return: An empty dictionary, as no specific state is stored.
        """
        return {}

    @classmethod
    def __unpacked__(cls, **kwargs: Any) -> "NotExistMatcher":
        """
        Unpack a NotExistMatcher instance from its serialized form.

        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of NotExistMatcher.
        """
        return cls()



from typing import Any, Dict, Union

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, DictOrTupleList, MultiDictMatcher
from ._request_matcher import RequestMatcher

__all__ = ("HeaderMatcher", "DictOrTupleListOrAttrMatcher",)


DictOrTupleListOrAttrMatcher = Union[DictOrTupleList, AttributeMatcher]


@packable("jj.matchers.HeaderMatcher")
class HeaderMatcher(RequestMatcher):
    """
    Matches HTTP requests based on headers.

    This matcher checks if the incoming request's headers match
    a specified set of headers, or a dynamic attribute-based matcher.
    """

    def __init__(self, headers: DictOrTupleListOrAttrMatcher, *, resolver: Resolver) -> None:
        """
        Initialize a HeaderMatcher with the specified headers and resolver.

        :param headers: The headers to match against, either as a dictionary,
                        list of tuples, or a dynamic attribute matcher.
        :param resolver: The resolver responsible for registering this matcher.
        """
        super().__init__(resolver=resolver)
        if isinstance(headers, AttributeMatcher):
            self._matcher = headers
        else:
            self._matcher = MultiDictMatcher(headers)

    @property
    def sub_matcher(self) -> AttributeMatcher:
        """
        Return the underlying attribute matcher used for header matching.

        :return: The matcher that evaluates the headers.
        """
        return self._matcher

    async def match(self, request: Request) -> bool:
        """
        Determine if the request's headers match the expected headers.

        :param request: The HTTP request containing the headers to match.
        :return: `True` if the request's headers match the expected values, otherwise `False`.
        """
        return await self._matcher.match(request.headers)

    def __repr__(self) -> str:
        """
        Return a string representation of the HeaderMatcher instance.

        :return: A string describing the class, matcher, and resolver.
        """
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the HeaderMatcher instance for serialization.

        :return: A dictionary containing the serialized form of the matcher.
        """
        return {"headers": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     headers: DictOrTupleListOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "HeaderMatcher":
        """
        Unpack a HeaderMatcher instance from its serialized form.

        :param headers: The headers matcher to use for this instance.
        :param resolver: The resolver to bind this matcher to.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of HeaderMatcher.
        """
        return cls(headers, resolver=resolver)



from inspect import getmembers
from typing import Any, Dict, Optional, Tuple, Type, Union

from ..handlers import HandlerFunction
from ..resolvers import Resolver
from ._abstract_app import AbstractApp

__all__ = ("BaseApp", "define_app", "create_app")


def _register(cls: Any, name: str, value: Any) -> None:
    if callable(value) and not name.startswith("_"):
        cls.resolver.register_handler(value, cls)


def _deregister(cls: Any, name: str) -> None:
    if hasattr(cls, name):
        cls.resolver.deregister_handler(getattr(cls, name), cls)


class _Registrar(type):
    def __init__(cls,
                 name: str,
                 bases: Tuple[Type[AbstractApp]],
                 namespace: Dict[str, Any]) -> None:
        for name, value in namespace.items():
            _register(cls, name, value)
        for base in bases:
            for name, member in getmembers(base):
                _register(cls, name, member)
        super().__init__(name, bases, namespace)

    def __setattr__(cls, name: str, value: Any) -> None:
        _register(cls, name, value)
        super().__setattr__(name, value)

    def __delattr__(cls, name: str) -> None:
        _deregister(cls, name)
        super().__delattr__(name)


class BaseApp(AbstractApp, metaclass=_Registrar):
    @property
    def resolver(self) -> Resolver:
        raise NotImplementedError()


def define_app(name: Optional[str] = None, *,
               resolver: Optional[Resolver] = None,
               handlers: Optional[Dict[str, HandlerFunction]] = None) -> Type[BaseApp]:
    if name is None:
        define_app.count = getattr(define_app, "count", 1) + 1  # type: ignore
        name = "App" + str(define_app.count)  # type: ignore

    namespace: Dict[str, Union[Resolver, HandlerFunction]] = {}
    if resolver:
        namespace["resolver"] = resolver
    if handlers:
        namespace.update(handlers)

    return type(name, (BaseApp,), namespace)


def create_app(name: Optional[str] = None, *,
               resolver: Optional[Resolver] = None,
               handlers: Optional[Dict[str, HandlerFunction]] = None) -> BaseApp:
    return define_app(name, resolver=resolver, handlers=handlers)()



from typing import List, Optional, Tuple, Union, cast

from aiohttp import ClientSession
from packed import pack, unpack

from jj import version
from jj.expiration_policy import ExpirationPolicy
from jj.http.codes import OK
from jj.http.methods import DELETE, GET, POST
from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import HistoryAdapterType, HistoryItem, default_history_adapter
from ._remote_handler import RemoteHandler
from ._remote_response import RemoteResponseType

__all__ = ("RemoteMock",)


class _RemoteMockError(Exception):
    pass


class RemoteMock:
    def __init__(self, url: str) -> None:
        self._url = url

    def create_handler(self,
                       matcher: Union[RequestMatcher, LogicalMatcher],
                       response: RemoteResponseType,
                       expiration_policy: Optional[ExpirationPolicy] = None,
                       *,
                       history_adapter: Optional[HistoryAdapterType] = default_history_adapter
                       ) -> RemoteHandler:
        return RemoteHandler(
            self,
            matcher,
            response,
            expiration_policy=expiration_policy,
            history_adapter=history_adapter,
        )

    async def _do_request(self, method: str, url: str,
                          data: Optional[bytes] = None) -> Tuple[int, bytes]:
        headers = {"x-jj-remote-mock": f"v{version}"}
        async with ClientSession() as session:
            async with session.request(method, url, data=data, headers=headers) as response:
                body = await response.read()
                return response.status, body

    def _pack_payload(self, handler: RemoteHandler) -> bytes:
        payload = {
            "id": str(handler.id),
            "request": handler.matcher,
            "response": handler.response,
            "expiration_policy": handler.expiration_policy,
        }
        try:
            return pack(payload)
        except Exception as e:
            error_message = (
                f"Error while packing payload: {e}. "
                "Make sure your payload is packable https://pypi.org/project/packed"
            )
            raise ValueError(error_message) from e

    async def register(self, handler: RemoteHandler) -> "RemoteMock":
        url = f"{self._url}/__jj__/register"
        status, body = await self._do_request(POST, url, self._pack_payload(handler))
        if status != OK:
            raise _RemoteMockError(f"Can't register mock ({body!r})")
        return self

    async def deregister(self, handler: RemoteHandler) -> "RemoteMock":
        url = f"{self._url}/__jj__/deregister"
        status, body = await self._do_request(DELETE, url, self._pack_payload(handler))
        if status != OK:
            raise _RemoteMockError(f"Can't deregister mock ({body!r})")
        return self

    async def fetch_history(self, handler: RemoteHandler) -> List[HistoryItem]:
        url = f"{self._url}/__jj__/history"
        status, body = await self._do_request(GET, url, self._pack_payload(handler))
        if status != OK:
            raise _RemoteMockError(f"Can't retrieve mock history ({body!r})")
        return cast(List[HistoryItem], unpack(body))

    async def reset(self) -> "RemoteMock":
        url = f"{self._url}/__jj__/reset"
        status, body = await self._do_request(POST, url)
        if status != OK:
            raise _RemoteMockError(f"Can't reset mock ({body!r})")
        return self



from ._body_parser import BodyParser
from ._history_adapter import HistoryAdapterType, default_history_adapter
from ._history_formatter import HistoryFormatter, PrettyHistoryFormatter
from ._history_item import HistoryItem
from ._history_repository import HistoryRepository
from ._history_request import HistoryRequest
from ._history_response import HistoryResponse

__all__ = ("HistoryRepository", "HistoryRequest", "HistoryResponse", "HistoryItem",
           "HistoryAdapterType", "default_history_adapter", "BodyParser",
           "HistoryFormatter", "PrettyHistoryFormatter",)



from typing import Union

from jj import DelayedResponse, RelayResponse, Response, TemplateResponse

REMOTE_RESPONSES = (Response, RelayResponse, DelayedResponse, TemplateResponse)

RemoteResponseType = Union[Response, RelayResponse, DelayedResponse, TemplateResponse]

__all__ = ("RemoteResponseType", "REMOTE_RESPONSES",)



from typing import Any

__all__ = ("AttributeMatcher",)


class AttributeMatcher:
    """
    Base class for matchers that match attributes of an HTTP request.

    This abstract class defines the interface for all matchers that evaluate
    various attributes (e.g., headers, query parameters) of a request.
    """

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value matches the criteria defined by the matcher.

        :param actual: The value to be matched against.
        :return: `True` if the value matches the criteria, otherwise `False`.
        :raises NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        """
        Return a string representation of the AttributeMatcher instance.

        :return: A string describing the class.
        """
        return f"{self.__class__.__qualname__}()"



from ._attribute_matcher import AttributeMatcher
from ._contain_matcher import ContainMatcher, NotContainMatcher
from ._equal_matcher import EqualMatcher, NotEqualMatcher
from ._exist_matcher import ExistMatcher, NotExistMatcher
from ._multi_dict_matcher import DictOrTupleList, MultiDictMatcher, StrOrAttrMatcher
from ._regex_matcher import RegexMatcher
from ._route_matcher import RouteMatcher

__all__ = (
    "AttributeMatcher",
    "ContainMatcher",
    "DictOrTupleList",
    "EqualMatcher",
    "ExistMatcher",
    "NotExistMatcher",
    "MultiDictMatcher",
    "NotContainMatcher",
    "NotEqualMatcher",
    "RegexMatcher",
    "RouteMatcher",
    "StrOrAttrMatcher",
)



from ._core import (
    App,
    Middleware,
    default_app,
    default_handler,
    default_logger,
    match,
    match_all,
    match_any,
    match_header,
    match_headers,
    match_method,
    match_methods,
    match_param,
    match_params,
    match_path,
    serve,
    start,
    wait_for,
)
from ._version import server_version, version
from .requests import Request
from .responses import (
    DelayedResponse,
    RelayResponse,
    Response,
    StaticResponse,
    StreamResponse,
    TemplateResponse,
)

__all__ = (
    "App",
    "Middleware",
    "match_method", "match_methods",
    "match_path",
    "match_headers", "match_header",
    "match_params", "match_param",
    "match_all", "match_any",
    "match",
    "start", "serve", "wait_for",
    "default_app", "default_handler", "default_logger",
    "version", "server_version",
    "Request",
    "Response", "StaticResponse", "StreamResponse", "RelayResponse",
    "DelayedResponse", "TemplateResponse",
)

__version__ = version



import asyncio
from asyncio import AbstractEventLoop, CancelledError, Event, Task, all_tasks, gather
from typing import Any, List, Optional, Sequence, Type

from aiohttp.web_runner import BaseRunner, TCPSite

__all__ = ("Server",)


class Server:
    def __init__(self, loop: AbstractEventLoop,
                 runner_factory: Type[BaseRunner],
                 site_factory: Type[TCPSite]) -> None:
        self._loop = loop
        self._runner_factory = runner_factory
        self._site_factory = site_factory
        self._tasks: List["Task[Any]"] = []
        self._events: List[Event] = []

    async def _start(self, app: Any, event: Event,
                     host: Optional[str] = None,
                     port: Optional[int] = None) -> None:
        runner = self._runner_factory(app, loop=self._loop)  # type: ignore
        await runner.setup()

        site = self._site_factory(runner, host=host, port=port)
        await site.start()

        try:
            await event.wait()
        except CancelledError:
            pass
        finally:
            await runner.cleanup()

    def start(self, app: Any, host: Optional[str] = None, port: Optional[int] = None) -> None:
        if self._loop is None:
            try:
                self._loop = asyncio.get_running_loop()
            except RuntimeError:
                self._loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self._loop)
        event = Event()
        self._events.append(event)
        task = self._loop.create_task(self._start(app, event, host, port))
        self._tasks.append(task)

    async def _serve(self) -> None:
        try:
            await gather(*self._tasks)
        finally:
            self.cleanup()

    def serve(self, exceptions: Sequence[Type[BaseException]] = (KeyboardInterrupt,)) -> None:
        try:
            self._loop.run_until_complete(self._serve())
        except tuple(exceptions):
            pass

    def cleanup(self) -> None:
        while self._events:
            event = self._events.pop()
            event.set()

    def _cancel_tasks(self, tasks: List["Task[Any]"]) -> None:
        if not tasks:
            return

        for task in tasks:
            task.cancel()

        self._loop.run_until_complete(gather(*tasks, return_exceptions=True))

        for task in tasks:
            if task.cancelled():
                continue
            if task.exception() is not None:
                self._loop.call_exception_handler({
                    "message": "unhandled exception during shutdown",
                    "exception": task.exception(),
                    "task": task,
                })

    def shutdown(self) -> None:
        try:
            self._cancel_tasks(list(all_tasks(self._loop)))
            self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        finally:
            self._loop.close()



import asyncio
import threading
from concurrent.futures import Future
from typing import Any, Callable, Coroutine, TypeVar

__all__ = ("Thread", "run_async",)

T = TypeVar("T")
TargetType = Callable[..., Coroutine[None, None, T]]


class Thread(threading.Thread):
    """
    Thread that runs an async function in a new event loop.

    Uses a thread-safe concurrent.futures.Future to communicate results
    between threads.
    """

    def __init__(self, future: "Future[T]", target: TargetType[T],
                 *args: Any, **kwargs: Any) -> None:
        """
        Initialize the thread with target async function and arguments.

        :param future: Thread-safe future to set the result on completion.
        :param target: Async function to run in the new event loop.
        :param args: Positional arguments for the target function.
        :param kwargs: Keyword arguments for the target function.
        """
        super().__init__()
        self._future = future
        self._target = target
        self._args = args
        self._kwargs = kwargs

    def run(self) -> None:
        """
        Run the async function in a new event loop and set the result.

        Creates a new event loop for this thread and safely sets the result
        using the thread-safe concurrent.futures.Future.
        """
        try:
            result = asyncio.run(self._target(*self._args, **self._kwargs))
        except Exception as e:
            self._future.set_exception(e)
        else:
            self._future.set_result(result)


def run_async(target: TargetType[T], *args: Any, **kwargs: Any) -> T:
    """
    Run an async function, handling both sync and async contexts safely.

    If called from a synchronous context (no running loop), runs the target
    directly. If called from an async context (with a running loop), runs
    the target in a separate thread with its own event loop to avoid blocking.

    :param target: Async function to execute.
    :param args: Positional arguments for the target function.
    :param kwargs: Keyword arguments for the target function.
    :return: The result of the async function.
    """
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.run(target(*args, **kwargs))
    else:
        future: "Future[T]" = Future()

        thread = Thread(future, target, *args, **kwargs)
        thread.start()
        thread.join()

        return future.result()



from ..handlers import HandlerFunction
from ._abstract_middleware import AbstractMiddleware
from ._middleware_type import MiddlewareType

__all__ = ("RootMiddleware",)


class RootMiddleware(AbstractMiddleware):
    def __call__(self, handler: HandlerFunction) -> MiddlewareType:
        return self._do



import logging
from logging import LogRecord

from ..requests import Request
from ..responses import Response

__all__ = ("Filter",)


class Filter(logging.Filter):
    def filter_request(self, request: Request, record: LogRecord) -> bool:
        return super().filter(record) is not False

    def filter_response(self, response: Response, request: Request, record: LogRecord) -> bool:
        return super().filter(record) is not False

    def filter(self, record: LogRecord) -> bool:
        request = getattr(record, "jj_request", None)
        response = getattr(record, "jj_response", None)
        if (response is not None) and (request is not None):
            return self.filter_response(response, request, record)
        elif request is not None:
            return self.filter_request(request, record)
        else:
            return super().filter(record) is not False



from typing import List, Type

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ._resolver import Resolver

__all__ = ("ReversedResolver",)


class ReversedResolver(Resolver):
    def get_handlers(self, app: Type[AbstractApp]) -> List[HandlerFunction]:
        handlers = super().get_handlers(app)
        return list(reversed(handlers))



from ..requests import Request
from ..responses import Response

__all__ = ("default_handler",)


async def default_handler(request: Request) -> Response:
    """
    Handle unmatched requests with a default 404 response.

    This asynchronous function is used as the default handler for HTTP requests
    that do not match any defined routes or handlers. It returns a `Response` object
    with a 404 status code, indicating that the requested resource was not found.

    :param request: The incoming HTTP request that triggered the default handler.
    :return: A `Response` object with a 404 status code.
    """
    return Response(status=404)



from abc import ABC, abstractmethod
from typing import List

from jj.mock._history._history_item import HistoryItem

__all__ = ("HistoryFormatter",)


class HistoryFormatter(ABC):
    @abstractmethod
    def format_history(self, history: List[HistoryItem]) -> str:
        pass



from datetime import datetime
from typing import List, TypedDict

from ._history_request import HistoryRequest
from ._history_response import HistoryResponse

__all__ = ("HistoryItem",)


class HistoryItem(TypedDict):
    request: HistoryRequest
    response: HistoryResponse
    tags: List[str]
    created_at: datetime



import os
from typing import Union

__all__ = (
    "get_remote_mock_url",
    "get_remote_mock_disposable",
    "get_remote_mock_pprint",
    "get_remote_mock_pprint_length",
    "get_remote_mock_pprint_width",
)


def get_remote_mock_url() -> str:
    """
    Get the remote mock URL from environment variable.

    This method dynamically fetches the value from the environment,
    avoiding race conditions with load_dotenv().

    :return: The remote mock URL, defaults to "http://localhost:8080".
    """
    return os.environ.get("JJ_REMOTE_MOCK_URL", "http://localhost:8080")


def get_remote_mock_disposable() -> bool:
    """
    Get the remote mock disposable setting from environment variable.

    This method dynamically fetches the value from the environment,
    avoiding race conditions with load_dotenv().

    :return: `True` if disposable mode is enabled, defaults to `True`.
    """
    value = os.environ.get("JJ_REMOTE_MOCK_DISPOSABLE", "True")
    return value.lower() in ("true", "1", "yes", "on")


def get_remote_mock_pprint() -> bool:
    """
    Get the remote mock pretty print setting from environment variable.

    This method dynamically fetches the value from the environment,
    avoiding race conditions with load_dotenv().

    :return: `True` if pretty printing is enabled, defaults to `False`.
    """
    value = os.environ.get("JJ_REMOTE_MOCK_PPRINT", "False")
    return value.lower() in ("true", "1", "yes", "on")


def get_remote_mock_pprint_length() -> Union[int, None]:
    """
    Get the remote mock pretty print length from environment variable.

    This method dynamically fetches the value from the environment,
    avoiding race conditions with load_dotenv().

    :return: The pprint length limit, or `None` if not set or invalid.
    """
    value = os.environ.get("JJ_REMOTE_MOCK_PPRINT_LENGTH", "")
    if value:
        try:
            return int(value)
        except ValueError:
            return None
    return None


def get_remote_mock_pprint_width() -> Union[int, None]:
    """
    Get the remote mock pretty print width from environment variable.

    This method dynamically fetches the value from the environment,
    avoiding race conditions with load_dotenv().

    :return: The pprint width limit, or `None` if not set or invalid.
    """
    value = os.environ.get("JJ_REMOTE_MOCK_PPRINT_WIDTH", "")
    if value:
        try:
            return int(value)
        except ValueError:
            return None
    return None



from http.cookies import Morsel
from typing import Any, Dict, Union

from aiohttp.payload import BytesPayload, IOBasePayload, TextIOPayload

__all__ = ("cookie_to_dict", "get_response_body",)


def cookie_to_dict(cookie: "Morsel[str]") -> Dict[str, Union[str, None]]:
    """
    Convert a Morsel object into a dictionary.

    :param cookie: The cookie morsel object to be converted.
    :return: A dictionary representation of the cookie, including attributes like
             "name", "value", "expires", "domain", "max-age", "path", "secure",
             "httponly", and "version".
    """
    dictionary: Dict[str, Union[str, None]] = {
        "name": cookie.key,
        "value": cookie.value,
    }
    for attr in ("expires", "domain", "max-age", "path", "secure", "httponly", "version"):
        key = attr.replace("-", "_")
        val = cookie.get(attr)
        dictionary[key] = None if val == "" else val
    return dictionary


def get_response_body(body: Any) -> bytes:
    """
    Convert the provided response body into bytes.

    :param body: The response body, which can be of types like bytes, bytearray,
                 memoryview, or specific aiohttp payloads.
    :return: The body converted to a bytes object.
    :raises ValueError: If the body type is unsupported.
    """
    if isinstance(body, (bytes, bytearray, memoryview)):
        return bytes(body)
    elif isinstance(body, BytesPayload):
        return bytes(body._value)
    elif isinstance(body, TextIOPayload):
        return bytes(body._value.read(), body.encoding)  # type: ignore
    elif isinstance(body, IOBasePayload):
        return bytes(body._value.read())
    else:
        raise ValueError("Unsupported body type {}".format(type(body)))



import json
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from aiohttp.typedefs import LooseHeaders
from aiohttp.web_request import BaseRequest
from multidict import CIMultiDict
from packed import packable

from .._version import server_version

try:
    import jinja2
except ImportError:  # pragma: no cover
    jinja2 = None  # type: ignore


from ._response import Response

__all__ = ("TemplateResponse",)


@packable("jj.responses.TemplateResponse")
class TemplateResponse(Response):
    """
    Represents an HTTP response where the body, headers, and status can be templated.

    The `TemplateResponse` class extends the `Response` class and allows for the use
    of Jinja2 templates to dynamically render the response body, headers, and status
    based on the request context. It relies on Jinja2 as an optional dependency.
    """

    def __init__(self, body: Optional[str] = None, *,
                 headers: Optional[LooseHeaders] = None,
                 status: Union[int, str] = 200,
                 context: Union[Any, None] = None) -> None:
        """
        Initialize the TemplateResponse with optional templated body, headers, status, and context.

        :param body: The template string for the response body. Defaults to an empty string.
        :param headers: A dictionary of headers, which can also contain templates.
        :param status: The HTTP status code, which can be templated as well (default is 200).
        :param context: The additional context data to be used in template rendering.
        """
        super().__init__()
        self._tmpl_body = body or ""
        self._tmpl_headers = CIMultiDict(headers or {})
        self._tmpl_status = str(status)
        self._tmpl_context = context
        self._jinja_env: Any = None
        self._prepare_hook_called = False

    async def _prepare_hook(self, request: BaseRequest) -> "TemplateResponse":
        """
        Prepare the response by rendering the templates for body, headers, and status.

        This method is called before sending the response to the client. It renders
        the templated body, headers, and status code using the request and context data.

        :param request: The incoming HTTP request object used for rendering templates.
        :return: The `TemplateResponse` object after template rendering is complete.
        :raises ImportError: If Jinja2 is not installed.
        """
        if self._prepare_hook_called:
            return self

        if jinja2 is None:  # pragma: no cover
            raise ImportError(
                "Jinja2 is an optional dependency. "
                "To use TemplateResponse, please install Jinja2 via 'pip install jinja2'"
            )
        self._jinja_env = jinja2.environment.Environment()
        # 'tojson' is a default filter provided by Jinja2
        self._jinja_env.filters["fromjson"] = json.loads

        context = {
            "request": request,
            "context": self._tmpl_context,
        }

        self._headers = CIMultiDict({
            "Server": self._tmpl_headers.get("Server", server_version)
        })
        self._headers.extend(
            (self._render_value(key, context), self._render_value(value, context))
            for key, value in self._tmpl_headers.items()
            if key.lower() != "server"
        )
        self.body = self._render_value(self._tmpl_body, context)  # type: ignore
        self.set_status(int(self._render_value(self._tmpl_status, context)))

        self._prepare_hook_called = True
        return self

    def _render_value(self, value: str, context: Dict[str, Any]) -> str:
        """
        Render a single template value using the request and context data.

        This method renders a given string value as a Jinja2 template, passing the
        request and additional context into the template rendering.

        :param value: The string value to be rendered as a template.
        :param context: The dictionary containing request and additional context data.
        :return: The rendered template as a string.
        """
        template = self._jinja_env.from_string(value)
        rendered = template.render(context)
        return cast(str, rendered)

    def copy(self) -> "TemplateResponse":
        """
        Create a copy of the TemplateResponse object.

        :return: A new instance of `TemplateResponse` with the same template configuration.
        :raises AssertionError: If the response is already prepared.
        """
        assert not self.prepared
        return self.__class__.__unpacked__(**self.__packed__())

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the TemplateResponse into a dictionary for serialization.

        This method prepares the response for serialization by packing its templated
        body, headers, status, and context into a dictionary.

        :return: A dictionary representing the packed TemplateResponse.
        """
        return {
            "body": self._tmpl_body,
            "headers": [
                [key, val] for key, val in self._tmpl_headers.items()
            ],
            "status": self._tmpl_status,
            "context": self._tmpl_context,
        }

    @classmethod
    def __unpacked__(cls, *,  # type: ignore
                     body: str,
                     headers: List[Tuple[str, str]],
                     status: Union[int, str],
                     context: Optional[Any] = None,
                     **kwargs: Any) -> "TemplateResponse":
        """
        Reconstruct a TemplateResponse instance from unpacked parameters.

        :param body: The template string for the response body.
        :param headers: A list of headers, each represented as a tuple of (key, value).
        :param status: The HTTP status code as a string or integer.
        :param context: The additional context data to be used in template rendering.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of `TemplateResponse`.
        """
        return cls(status=status, headers=headers, body=body, context=context)



from os import linesep
from pprint import pformat
from shutil import get_terminal_size
from typing import List, Optional

from .._history_item import HistoryItem
from ._history_formatter import HistoryFormatter

__all__ = ("PrettyHistoryFormatter",)


class PrettyHistoryFormatter(HistoryFormatter):
    def __init__(self, width: Optional[int] = None, length: Optional[int] = None) -> None:
        self._width = width if width else self._get_terminal_width()
        self._length = length

    def _get_terminal_width(self) -> int:
        columns, _ = get_terminal_size()
        return columns

    def _cut_str(self, string: str, length: int, separator: str = "...") -> str:
        assert length > len(separator)
        if len(string) <= length:
            return string
        length -= len(separator)
        return string[:length // 2] + separator + string[-length // 2:]

    def format_history(self, history: List[HistoryItem]) -> str:
        result = []
        for index, history_item in enumerate(history):
            formatted = pformat(
                {
                    "index": index,
                    "request": history_item["request"].to_dict(),
                    "response": history_item["response"].to_dict(),
                    # "tags": history_item["tags"],
                },
                width=self._width
            )
            if self._length:
                formatted = self._cut_str(formatted, self._length)
            result.append(formatted)
        if len(result) == 0:
            return "[]"
        return f"[{linesep}" + f",{linesep}".join(result) + f"{linesep}]"



from ..requests import Request

__all__ = ("ExpirationPolicy",)


class ExpirationPolicy:
    """
    Base class for defining expiration policies in HTTP mocking.

    An expiration policy determines if a mocked response should still be
    considered valid based on the criteria defined by its implementation.
    """

    async def is_expired(self, request: Request) -> bool:
        """
        Determine whether the response associated with this policy has expired.

        :param request: The request to evaluate against the expiration criteria.
        :return: `True` if the policy deems the response expired, otherwise `False`.
        :raises NotImplementedError: If this method is not implemented by a subclass.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        """
        Return a string representation of the ExpirationPolicy instance.

        :return: A string representation of the instance.
        """
        return f"{self.__class__.__qualname__}()"



from typing import Any, Dict

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from ..attribute_matchers import AttributeMatcher, EqualMatcher, StrOrAttrMatcher
from ._request_matcher import RequestMatcher

__all__ = ("MethodMatcher",)


@packable("jj.matchers.MethodMatcher")
class MethodMatcher(RequestMatcher):
    """
    Matches HTTP requests based on the HTTP method.

    This matcher checks if the incoming request uses a specified HTTP method
    (e.g., GET, POST). It supports matching against a static method or an
    attribute-based matcher.
    """

    def __init__(self, method: StrOrAttrMatcher, *, resolver: Resolver) -> None:
        """
        Initialize a MethodMatcher with the HTTP method and resolver.

        :param method: The HTTP method to match against, or a matcher that can
                       evaluate the method dynamically.
        :param resolver: The resolver responsible for registering this matcher.
        """
        super().__init__(resolver=resolver)
        if isinstance(method, AttributeMatcher):
            self._matcher = method
        else:
            self._matcher = EqualMatcher(str.upper(method))

    @property
    def sub_matcher(self) -> AttributeMatcher:
        """
        Return the underlying attribute matcher used for the method matching.

        :return: The matcher that evaluates the method condition.
        """
        return self._matcher

    async def match(self, request: Request) -> bool:
        """
        Determine if the request method matches the expected method.

        :param request: The HTTP request containing the method to match.
        :return: `True` if the request method matches the expected method, otherwise `False`.
        """
        return await self._matcher.match("*") or await self._matcher.match(request.method)

    def __repr__(self) -> str:
        """
        Return a string representation of the MethodMatcher instance.

        :return: A string describing the class, matcher, and resolver.
        """
        return f"{self.__class__.__qualname__}({self._matcher!r}, resolver={self._resolver!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the MethodMatcher instance for serialization.

        :return: A dictionary containing the serialized form of the matcher.
        """
        return {"method": self._matcher}

    @classmethod
    def __unpacked__(cls, *,
                     method: StrOrAttrMatcher,
                     resolver: Resolver,
                     **kwargs: Any) -> "MethodMatcher":
        """
        Unpack a MethodMatcher instance from its serialized form.

        :param method: The method matcher to use for this instance.
        :param resolver: The resolver to bind this matcher to.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of MethodMatcher.
        """
        return cls(method, resolver=resolver)



from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("EqualMatcher", "NotEqualMatcher",)


@packable("jj.matchers.EqualMatcher")
class EqualMatcher(AttributeMatcher):
    """
    Matches if the actual value equals the expected value.

    This matcher evaluates whether the actual value in a request matches
    the predefined expected value.
    """

    def __init__(self, expected: Any) -> None:
        """
        Initialize an EqualMatcher with an expected value.

        :param expected: The value to compare against the actual value.
        """
        self._expected = expected

    @property
    def expected(self) -> Any:
        """
        Return the expected value for this matcher.

        :return: The expected value.
        """
        return self._expected

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value matches the expected value.

        :param actual: The value to compare against the expected value.
        :return: `True` if the actual value equals the expected value, otherwise `False`.
        """
        return bool(self._expected == actual)

    def __repr__(self) -> str:
        """
        Return a string representation of the EqualMatcher instance.

        :return: A string describing the class and expected value.
        """
        return f"{self.__class__.__qualname__}({self._expected!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the EqualMatcher instance for serialization.

        :return: A dictionary containing the expected value.
        """
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "EqualMatcher":
        """
        Unpack an EqualMatcher instance from its serialized form.

        :param expected: The expected value to use for the matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of EqualMatcher.
        """
        return cls(expected)


@packable("jj.matchers.NotEqualMatcher")
class NotEqualMatcher(EqualMatcher):
    """
    Matches if the actual value does not equal the expected value.

    This matcher evaluates whether the actual value in a request does not
    match the predefined expected value.
    """

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value does not match the expected value.

        :param actual: The value to compare against the expected value.
        :return: `True` if the actual value does not equal the expected value, otherwise `False`.
        """
        return bool(self._expected != actual)

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the NotEqualMatcher instance for serialization.

        :return: A dictionary containing the expected value.
        """
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "NotEqualMatcher":
        """
        Unpack a NotEqualMatcher instance from its serialized form.

        :param expected: The expected value to use for the matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of NotEqualMatcher.
        """
        return cls(expected)



from inspect import signature

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ..requests import Request
from ..responses import StreamResponse
from ._root_middleware import RootMiddleware

__all__ = ("SelfMiddleware",)


class SelfMiddleware(RootMiddleware):
    async def do(self, request: Request, handler: HandlerFunction,
                 app: AbstractApp) -> StreamResponse:
        unwrapped = self._resolver.unwrap(handler)
        sig = signature(unwrapped)
        if len(sig.parameters) == 2:
            response = await handler(app, request)
        else:
            response = await handler(request)
        return response



from typing import Any, Dict

from ._abstract_app import AbstractApp

__all__ = ("DefaultApp",)


class _Singleton(type):
    _instances: Dict[Any, "_Singleton"] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> "_Singleton":
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class DefaultApp(AbstractApp, metaclass=_Singleton):
    pass



from typing import Any, Dict

from packed import packable

from ._attribute_matcher import AttributeMatcher

__all__ = ("ContainMatcher", "NotContainMatcher",)


@packable("jj.matchers.ContainMatcher")
class ContainMatcher(AttributeMatcher):
    """
    Matches if the actual value contains the expected value.

    This matcher checks if the expected value is present in the actual value.
    """

    def __init__(self, expected: Any) -> None:
        """
        Initialize a ContainMatcher with an expected value.

        :param expected: The value to check for in the actual value.
        """
        self._expected = expected

    @property
    def expected(self) -> Any:
        """
        Return the expected value for this matcher.

        :return: The expected value.
        """
        return self._expected

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value contains the expected value.

        :param actual: The value to check.
        :return: `True` if the actual value contains the expected value, otherwise `False`.
        """
        return self._expected in actual

    def __repr__(self) -> str:
        """
        Return a string representation of the ContainMatcher instance.

        :return: A string describing the class and expected value.
        """
        return f"{self.__class__.__qualname__}({self._expected!r})"

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the ContainMatcher instance for serialization.

        :return: A dictionary containing the expected value.
        """
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "ContainMatcher":
        """
        Unpack a ContainMatcher instance from its serialized form.

        :param expected: The expected value to use for this matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of ContainMatcher.
        """
        return cls(expected)


@packable("jj.matchers.NotContainMatcher")
class NotContainMatcher(ContainMatcher):
    """
    Matches if the actual value does not contain the expected value.

    This matcher checks if the expected value is not present in the actual value.
    """

    async def match(self, actual: Any) -> bool:
        """
        Determine if the actual value does not contain the expected value.

        :param actual: The value to check.
        :return: `True` if the actual value does not contain the expected value, otherwise `False`.
        """
        return self._expected not in actual

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the NotContainMatcher instance for serialization.

        :return: A dictionary containing the expected value.
        """
        return {"expected": self._expected}

    @classmethod
    def __unpacked__(cls, *, expected: Any, **kwargs: Any) -> "NotContainMatcher":
        """
        Unpack a NotContainMatcher instance from its serialized form.

        :param expected: The expected value to use for this matcher.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of NotContainMatcher.
        """
        return cls(expected)



from .._resolvable_matcher import ResolvableMatcher

__all__ = ("RequestMatcher",)


class RequestMatcher(ResolvableMatcher):
    """
    Serves as a base class for specific HTTP request matchers.

    Inherits from `ResolvableMatcher` and is intended to be subclassed by
    matchers that deal with different aspects of an HTTP request, such as
    method, path, headers, etc.
    """
    pass



from typing import Any, Dict, List

from packed import packable

from ...requests import Request
from ...resolvers import Resolver
from .._resolvable_matcher import ResolvableMatcher
from ._logical_matcher import LogicalMatcher

__all__ = ("AnyMatcher",)


@packable("jj.matchers.AnyMatcher")
class AnyMatcher(LogicalMatcher):
    """
    Matches an HTTP request if any one of the provided matchers succeeds.

    This matcher combines multiple `ResolvableMatcher` instances, and it
    returns `True` as soon as one of them successfully matches the incoming
    HTTP request.
    """

    def __init__(self, matchers: List[ResolvableMatcher], *, resolver: Resolver) -> None:
        """
        Initialize an AnyMatcher with a list of matchers and a resolver.

        :param matchers: A list of matchers to evaluate. Matching succeeds if at least
                         one matcher in this list returns `True`.
        :param resolver: The resolver responsible for registering this matcher.
        :raises AssertionError: If the matchers list is empty.
        """
        super().__init__(resolver=resolver)
        assert len(matchers) > 0
        self._matchers = matchers

    @property
    def sub_matchers(self) -> List[ResolvableMatcher]:
        """
        Return the list of matchers used by this `AnyMatcher`.

        :return: A copy of the list of matchers.
        """
        return self._matchers[:]

    async def match(self, request: Request) -> bool:
        """
        Determine if any matcher in the list matches the given request.

        :param request: The HTTP request to evaluate.
        :return: `True` if any of the matchers in the list returns `True`, otherwise `False`.
        """
        for matcher in self._matchers:
            if await matcher.match(request):
                return True
        return False

    def __repr__(self) -> str:
        """
        Return a string representation of the AnyMatcher instance.

        :return: A string describing the class, matchers, and resolver.
        """
        return (f"{self.__class__.__qualname__}"
                f"({self._matchers!r}, resolver={self._resolver!r})")

    def __packed__(self) -> Dict[str, Any]:
        """
        Pack the AnyMatcher instance for serialization.

        :return: A dictionary containing the serialized matchers.
        """
        return {"matchers": self._matchers}

    @classmethod
    def __unpacked__(cls, *,
                     matchers: List[ResolvableMatcher],
                     resolver: Resolver,
                     **kwargs: Any) -> "AnyMatcher":
        """
        Unpack an AnyMatcher instance from its serialized form.

        :param matchers: The list of matchers to use for this instance.
        :param resolver: The resolver to bind this matcher to.
        :param kwargs: Additional keyword arguments (ignored).
        :return: A new instance of AnyMatcher.
        """
        return cls(matchers, resolver=resolver)



from functools import partial
from typing import List, Optional, Sequence, Type

from aiohttp import web

from .apps import AbstractApp, BaseApp, DefaultApp
from .handlers import default_handler
from .logs import Logger
from .logs import logger as default_logger
from .matchers import ResolvableMatcher
from .matchers.logical_matchers import AllMatcher, AnyMatcher
from .matchers.request_matchers import (
    DictOrTupleListOrAttrMatcher,
    HeaderMatcher,
    MethodMatcher,
    ParamMatcher,
    PathMatcher,
    StrOrAttrMatcher,
)
from .middlewares import BaseMiddleware, LoggerMiddleware, SelfMiddleware
from .resolvers import Registry, ReversedResolver
from .runners import AppRunner
from .servers import Server

__all__ = (
    "App",
    "Middleware",
    "match_method", "match_methods",
    "match_path",
    "match_headers", "match_header",
    "match_params", "match_param",
    "match_all", "match_any",
    "match",
    "start", "wait_for", "serve",
    "default_app", "default_handler", "default_logger",
)

default_app = DefaultApp()
registry = Registry()
resolver = ReversedResolver(registry, default_app, default_handler)

runner = partial(AppRunner, resolver=resolver, handle_signals=True, middlewares=[
    SelfMiddleware(resolver),
])
# ignore because of https://github.com/python/mypy/issues/1484
server = Server(None, runner, web.TCPSite)  # type: ignore


class App(BaseApp):
    """
    Represents the main application class, extending `BaseApp`.

    This class uses the global `resolver` for request handling and matcher resolution.
    """
    resolver = resolver


class Middleware(BaseMiddleware):
    """
    Represents middleware used in the application, extending `BaseMiddleware`.

    This middleware also uses the global `resolver` for matcher resolution and handling.
    """
    resolver = resolver


def match_method(method: StrOrAttrMatcher) -> MethodMatcher:
    """
    Match an HTTP request based on its method.

    :param method: The HTTP method to match (e.g., "GET", "POST").
    :return: A `MethodMatcher` that will match requests with the given method.
    """
    return MethodMatcher(method, resolver=resolver)


def match_methods(*methods: StrOrAttrMatcher) -> AnyMatcher:
    """
    Match an HTTP request against multiple possible methods.

    :param methods: A variable number of HTTP methods to match.
    :return: An `AnyMatcher` that will match if any of the methods match.
    """
    matchers: List[ResolvableMatcher] = [
        MethodMatcher(method, resolver=resolver)
        for method in methods
    ]
    return AnyMatcher(matchers, resolver=resolver)


def match_path(path: StrOrAttrMatcher) -> PathMatcher:
    """
    Match an HTTP request based on the request path.

    :param path: The path or a matcher for the path to match (e.g., "/users").
    :return: A `PathMatcher` that will match requests with the given path.
    """
    return PathMatcher(path, resolver=resolver)


def match_headers(headers: DictOrTupleListOrAttrMatcher) -> HeaderMatcher:
    """
    Match an HTTP request based on its headers.

    :param headers: The headers or matchers for the headers to match.
    :return: A `HeaderMatcher` that will match requests with the given headers.
    """
    return HeaderMatcher(headers, resolver=resolver)


def match_header(name: str, value: StrOrAttrMatcher) -> HeaderMatcher:
    """
    Match an HTTP request based on a specific header.

    :param name: The name of the header to match.
    :param value: The expected value of the header or a matcher for the value.
    :return: A `HeaderMatcher` that will match requests with the specified header.
    """
    return HeaderMatcher([(name, value)], resolver=resolver)


def match_params(params: DictOrTupleListOrAttrMatcher) -> ParamMatcher:
    """
    Match an HTTP request based on its query parameters.

    :param params: The query parameters or matchers for the parameters.
    :return: A `ParamMatcher` that will match requests with the given query parameters.
    """
    return ParamMatcher(params, resolver=resolver)


def match_param(name: str, value: StrOrAttrMatcher) -> ParamMatcher:
    """
    Match an HTTP request based on a specific query parameter.

    :param name: The name of the query parameter to match.
    :param value: The expected value of the parameter or a matcher for the value.
    :return: A `ParamMatcher` that will match requests with the specified query parameter.
    """
    return ParamMatcher([(name, value)], resolver=resolver)


def match_all(matchers: List[ResolvableMatcher]) -> AllMatcher:
    """
    Match an HTTP request if all provided matchers succeed.

    :param matchers: A list of matchers that must all succeed for the request to match.
    :return: An `AllMatcher` that will match only if all the matchers succeed.
    """
    return AllMatcher(matchers, resolver=resolver)


def match_any(matchers: List[ResolvableMatcher]) -> AnyMatcher:
    """
    Match an HTTP request if any of the provided matchers succeed.

    :param matchers: A list of matchers where at least one must succeed for the request to match.
    :return: An `AnyMatcher` that will match if any of the matchers succeed.
    """
    return AnyMatcher(matchers, resolver=resolver)


def match(method: Optional[StrOrAttrMatcher] = None,
          path: Optional[StrOrAttrMatcher] = None,
          params: Optional[DictOrTupleListOrAttrMatcher] = None,
          headers: Optional[DictOrTupleListOrAttrMatcher] = None) -> AllMatcher:
    """
    Match an HTTP request based on multiple criteria such as method, path,
    query parameters, and headers.

    :param method: The HTTP method to match (optional).
    :param path: The request path to match (optional).
    :param params: The query parameters to match (optional).
    :param headers: The headers to match (optional).
    :return: An `AllMatcher` that matches if all specified conditions are met.
    """
    submatchers: List[ResolvableMatcher] = []
    if method:
        submatchers += [MethodMatcher(method, resolver=resolver)]
    if path:
        submatchers += [PathMatcher(path, resolver=resolver)]
    if params:
        submatchers += [ParamMatcher(params, resolver=resolver)]
    if headers:
        submatchers += [HeaderMatcher(headers, resolver=resolver)]
    return AllMatcher(submatchers, resolver=resolver)


def start(app: AbstractApp, *,
          host: Optional[str] = None, port: Optional[int] = None,
          logger: Logger = default_logger) -> None:
    """
    Start the HTTP server for the provided app.

    :param app: The application instance to run.
    :param host: The host address to bind to (optional).
    :param port: The port to bind to (optional).
    :param logger: The logger instance to use for logging (default: `default_logger`).
    """
    LoggerMiddleware(resolver, logger)(type(app))
    server.start(app, host=host, port=port)


def wait_for(exceptions: Sequence[Type[BaseException]]) -> None:
    """
    Wait for server termination, handling the provided exceptions.

    This method blocks until the server shuts down, listening for specified exceptions.

    :param exceptions: A sequence of exception types that should be handled to stop the server.
    """
    try:
        server.serve(exceptions)
    finally:
        server.shutdown()


def serve(app: AbstractApp = default_app, *,
          host: Optional[str] = None, port: Optional[int] = None,
          logger: Logger = default_logger) -> None:
    """
    Serve the provided app and wait for its termination.

    This method starts the app and blocks until the server is interrupted
    (e.g., by a keyboard interrupt).

    :param app: The application instance to run (default: `default_app`).
    :param host: The host address to bind to (optional).
    :param port: The port to bind to (optional).
    :param logger: The logger instance to use for logging (default: `default_logger`).
    """
    start(app, host=host, port=port, logger=logger)
    wait_for([KeyboardInterrupt])
